from six import iteritems, itervalues
from six.moves import range

from ..interfaces import ISnapshotable
from ..errors import AccessViolationError, InvalidResourceError
from ..util import align, sizeof_fmt, Flags
from ..snapshot import SnapshotNode

import enum

# Types
from ctypes import c_byte as i8_t     # NOQA
from ctypes import c_short as i16_t   # NOQA
from ctypes import c_int as i32_t     # NOQA
from ctypes import c_int64 as i64_t   # NOQA

from ctypes import c_ubyte as u8_t    # NOQA
from ctypes import c_ushort as u16_t  # NOQA
from ctypes import c_uint as u32_t    # NOQA
from ctypes import c_uint64 as u64_t  # NOQA

WORD_SIZE  = 4
SHORT_SIZE = 2

PAGE_SHIFT = 8
#: Size of memory page, in bytes.
PAGE_SIZE = (1 << PAGE_SHIFT)
PAGE_MASK = (~(PAGE_SIZE - 1))

MINIMAL_SIZE = 16

class MMOperationList(enum.IntEnum):
  ALLOC    = 3
  FREE     = 4
  UNUSED   = 5
  MMAP     = 6
  UNMMAP   = 7

from ..util import UINT8_FMT, UINT16_FMT, UINT32_FMT, UINT64_FMT  # noqa

def SIZE_FMT(size):
  return str(size)

def OFFSET_FMT(offset):
  s = '-' if offset < 0 else ''

  return '{}0x{:04X}'.format(s, abs(offset))

class MalformedBinaryError(Exception):
  pass

def addr_to_page(addr):
  return (addr & PAGE_MASK) >> PAGE_SHIFT

def addr_to_offset(addr):
  return addr & (PAGE_SIZE - 1)

def area_to_pages(addr, size):
  return ((addr & PAGE_MASK) >> PAGE_SHIFT, align(PAGE_SIZE, size) // PAGE_SIZE)

class PageTableEntry(Flags):
  _flags = ['read', 'write', 'execute', 'dirty']
  _labels = 'RWXD'

  READ    = 0x01
  WRITE   = 0x02
  EXECUTE = 0x04
  DIRTY   = 0x08

class MemoryPageState(SnapshotNode):
  def __init__(self, *args, **kwargs):
    super(MemoryPageState, self).__init__('index', 'content')

class MemoryPage(object):
  """
  Base class for all memory pages of any kinds.

  Memory page has a set of boolean flags that determine access to and behavior
  of the page.

  +-------------+-----------------------------------------------------------------------------+-----------+
  | Flag        | Meaning                                                                     | Default   |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``read``    | page is readable by executed instructions                                   | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``write``   | page is writable by executed instructions                                   | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``execute`` | content of the page can be used as executable instructions                  | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+
  | ``dirty``   | there have been write access to this page, its content has changed          | ``False`` |
  +-------------+-----------------------------------------------------------------------------+-----------+

  :param ducky.mm.MemoryController controller: Controller that owns this page.
  :param int index: Serial number of this page.
  """

  def __init__(self, controller, index):
    super(MemoryPage, self).__init__()

    self.controller = controller
    self.index = index

    self.DEBUG = self.controller.DEBUG
    self.INFO = self.controller.INFO
    self.WARN = self.controller.WARN
    self.ERROR = self.controller.ERROR
    self.EXCEPTION = self.controller.EXCEPTION

    self.base_address = self.index * PAGE_SIZE

  def __repr__(self):
    return '<%s index=%i, base=%s>' % (self.__class__.__name__, self.index, UINT32_FMT(self.base_address))

  def save_state(self, parent):
    """
    Create state of this page, and attach it to snapshot tree.

    :param parent: Parent snapshot node.
    :type parent: ducky.snapshot.SnapshotNode
    """

    state = parent.add_child('page_{}'.format(self.index), MemoryPageState())

    state.index = self.index
    state.content = [ord(i) if isinstance(i, str) else i for i in self.data]

    return state

  def load_state(self, state):
    """
    Restore page from a snapshot.
    """

    for i in range(0, PAGE_SIZE):
      self.data[i] = state.content[i]

  def __len__(self):
    """
    :return: length of this page. By default, all pages have the same length.
    :rtype: int
    """

    return PAGE_SIZE

  def clear(self):
    """
    Clear page.

    This operation is implemented by child classes.
    """

    raise NotImplementedError('Not allowed to clear memory on this address: page={}'.format(self.index))

  def read_u8(self, offset):
    """
    Read byte.

    This operation is implemented by child classes.

    :param int offset: offset of requested byte.
    :rtype: int
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

  def read_u16(self, offset):
    """
    Read word.

    This operation is implemented by child classes.

    :param int offset: offset of requested word.
    :rtype: int
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

  def read_u32(self, offset):
    """
    Read longword.

    This operation is implemented by child classes.

    :param int offset: offset of requested longword.
    :rtype: int
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

  def write_u8(self, offset, value):
    """
    Write byte.

    This operation is implemented by child classes.

    :param int offset: offset of requested byte.
    :param int value: value to write into memory.
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

  def write_u16(self, offset, value):
    """
    Write word.

    This operation is implemented by child classes.

    :param int offset: offset of requested word.
    :param int value: value to write into memory.
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

  def write_u32(self, offset, value):
    """
    Write longword.

    This operation is implemented by child classes.

    :param int offset: offset of requested longword.
    :param int value: value to write into memory.
    """

    raise NotImplementedError('Not allowed to access memory on this address: page={}, offset={}'.format(self.index, offset))

class AnonymousMemoryPage(MemoryPage):
  """
  "Anonymous" memory page - this page is just a plain array of bytes, and is
  not backed by any storage. Its content lives only in the memory.

  Page is created with all bytes set to zero.
  """

  def __init__(self, controller, index):
    super(AnonymousMemoryPage, self).__init__(controller, index)

    self.data = bytearray([0 for _ in range(0, PAGE_SIZE)])

  def clear(self):
    self.DEBUG('%s.clear', self.__class__.__name__)

    for i in range(0, PAGE_SIZE):
      self.data[i] = 0

  def read_u8(self, offset):
    self.DEBUG('%s.read_u8: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.data[offset]

  def read_u16(self, offset):
    self.DEBUG('%s.read_u16: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.data[offset] | (self.data[offset + 1] << 8)

  def read_u32(self, offset):
    self.DEBUG('%s.do_read_u32: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.data[offset] | (self.data[offset + 1] << 8) | (self.data[offset + 2] << 16) | (self.data[offset + 3] << 24)

  def write_u8(self, offset, value):
    self.DEBUG('%s.do_write_u8: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.data[offset] = value

  def write_u16(self, offset, value):
    self.DEBUG('%s.write_u16: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.data[offset]     =  value & 0x00FF
    self.data[offset + 1] = (value & 0xFF00) >> 8

  def write_u32(self, offset, value):
    self.DEBUG('%s.write_u32: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.data[offset]     =  value &       0xFF
    self.data[offset + 1] = (value &     0xFF00) >> 8
    self.data[offset + 2] = (value &   0xFF0000) >> 16
    self.data[offset + 3] = (value & 0xFF000000) >> 24

class ExternalMemoryPage(MemoryPage):
  """
  Memory page backed by an external source. Source is an array of bytes,
  and can be provided by device driver, mmaped file, or by any other mean.
  """

  def __init__(self, controller, index, data, offset = 0):
    super(ExternalMemoryPage, self).__init__(controller, index)

    self.data = data
    self.offset = offset

  def __repr__(self):
    return '<%s index=%i, base=%s, offset=%s>' % (self.__class__.__name__, self.index, UINT32_FMT(self.base_address), UINT32_FMT(self.offset))

  def save_state(self, parent):
    state = super(ExternalMemoryPage, self).save_state(parent)

    state.content = [ord(i) if isinstance(i, str) else i for i in self.data[self.offset:self.offset + PAGE_SIZE]]

  def clear(self):
    self.DEBUG('%s.clear', self.__class__.__name__)

    for i in range(0, PAGE_SIZE):
      self.data[i] = 0

  def get(self, offset):
    """
    Get one byte from page. Override this method in case you need a different
    offset of requested byte.

    :param int offset: offset of the requested byte.
    :rtype: int
    :returns: byte at position in page.
    """

    return self.data[self.offset + offset]

  def put(self, offset, b):
    """
    Put one byte into page. Override this method in case you need a different
    offset of requested byte.

    :param int offset: offset of modified byte.
    :param int b: new value.
    """

    self.data[self.offset + offset] = b

  def read_u8(self, offset):
    self.DEBUG('%s.read_u8: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset)

  def read_u16(self, offset):
    self.DEBUG('%s.read_u16: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset) | (self.get(offset + 1) << 8)

  def read_u32(self, offset):
    self.DEBUG('%s.read_u32: page=%s, offset=%s', self.__class__.__name__, self.index, offset)

    return self.get(offset) | (self.get(offset + 1) << 8) | (self.get(offset + 2) << 16) | (self.get(offset + 3) << 24)

  def write_u8(self, offset, value):
    self.DEBUG('%s.write_u8: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value)

  def write_u16(self, offset, value):
    self.DEBUG('%s.write_u16: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value & 0x00FF)
    self.put(offset + 1, (value & 0xFF00) >> 8)

  def write_u32(self, offset, value):
    self.DEBUG('%s.write_u32: page=%s, offset=%s, value=%s', self.__class__.__name__, self.index, offset, value)

    self.put(offset, value & 0x00FF)
    self.put(offset + 1, (value & 0xFF00) >> 8)
    self.put(offset + 2, (value & 0xFF0000) >> 16)
    self.put(offset + 3, (value & 0xFF000000) >> 24)

class MemoryRegionState(SnapshotNode):
  def __init__(self):
    super(MemoryRegionState, self).__init__('name', 'address', 'size', 'flags', 'pages_start', 'pages_cnt')

class MemoryRegion(ISnapshotable, object):
  region_id = 0

  def __init__(self, mc, name, address, size, flags):
    super(MemoryRegion, self).__init__()

    self.memory = mc

    self.id = MemoryRegion.region_id
    MemoryRegion.region_id += 1

    self.name = name
    self.address = address
    self.size = size
    self.flags = flags

    self.pages_start, self.pages_cnt = area_to_pages(self.address, self.size)

    self.memory.machine.DEBUG('MemoryRegion: name=%s, address=%s, size=%s, flags=%s, pages_start=%s, pages_cnt=%s', name, address, size, self.flags.to_string(), self.pages_start, self.pages_cnt)

  def __repr__(self):
    return '<MemoryRegion: name=%s, address=%s, size=%s, flags=%s, pages_start=%s, pages_cnt=%s' % (self.name, self.address, self.size, self.flags.to_string(), self.pages_start, self.pages_cnt)

  def save_state(self, parent):
    state = parent.add_child('memory_region_{}'.format(self.id), MemoryRegionState())

    state.name = self.name
    state.address = self.address
    state.size = self.size
    state.flags = self.flags.to_int()
    state.pages_start = self.pages_start
    state.pages_cnt = self.pages_cnt

  def load_state(self, state):
    pass

class MemoryState(SnapshotNode):
  def __init__(self):
    super(MemoryState, self).__init__('size')

  def get_page_states(self):
    return [__state for __name, __state in iteritems(self.get_children()) if __name.startswith('page_')]

class MemoryController(object):
  """
  Memory controller handles all operations regarding main memory.

  :param ducky.machine.Machine machine: virtual machine that owns this controller.
  :param int size: size of memory, in bytes.
  :raises ducky.errors.InvalidResourceError: when memory size is not multiple of
    :py:data:`ducky.mm.PAGE_SIZE`.
  """

  def __init__(self, machine, size = 0x1000000):
    machine.DEBUG('%s: size=0x%X', self.__class__.__name__, size)

    if size % PAGE_SIZE != 0:
      raise InvalidResourceError('Memory size must be multiple of PAGE_SIZE')

    if size < MINIMAL_SIZE * PAGE_SIZE:
      raise InvalidResourceError('Memory size must be at least %d pages' % MINIMAL_SIZE)

    self.machine = machine

    # Setup logging - create our local shortcuts to machine' logger
    self.DEBUG = self.machine.DEBUG
    self.INFO = self.machine.INFO
    self.WARN = self.machine.WARN
    self.ERROR = self.machine.ERROR
    self.EXCEPTION = self.machine.EXCEPTION

    self.force_aligned_access = self.machine.config.getbool('memory', 'force-aligned-access', default = False)

    self.size = size
    self.pages_cnt = size // PAGE_SIZE
    self.pages = {}

  def save_state(self, parent):
    self.DEBUG('mc.save_state')

    state = parent.add_child('memory', MemoryState())

    state.size = self.size

    for page in itervalues(self.pages):
      page.save_state(state)

  def load_state(self, state):
    self.size = state.size

    for page_state in state.get_children():
      page = self.get_page(page_state.index)
      page.load_state(page_state)

  def __set_page(self, pg):
    """
    Install page object for a specific memory page.

    :param ducky.mm.MemoryPage pg: page to be installed
    :returns: installed page
    :rtype: :py:class:`ducky.mm.MemoryPage`
    """

    assert pg.index not in self.pages

    if pg.index >= self.pages_cnt:
      raise InvalidResourceError('Attempt to create page with index out of bounds: pg.index=%d' % pg.index)

    self.pages[pg.index] = pg
    return pg

  def __remove_page(self, pg):
    """
    Removes page object for a specific memory page.

    :param ducky.mm.MemoryPage pg: page to be removed
    """

    assert pg.index in self.pages

    del self.pages[pg.index]

  def __alloc_page(self, index):
    """
    Allocate new anonymous page for usage. The first available index is used.

    Be aware that this method does NOT check if page is already allocated. If
    it is, it is just overwritten by new anonymous page.

    :param int index: index of requested page.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    """

    return self.__set_page(AnonymousMemoryPage(self, index))

  def alloc_specific_page(self, index):
    """
    Allocate new anonymous page with specific index for usage.

    :param int index: allocate page with this particular index.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.AccessViolationError: when page is already allocated.
    """

    self.DEBUG('mc.alloc_specific_page: index=%s', index)

    if index in self.pages:
      raise AccessViolationError('Page {} is already allocated'.format(index))

    return self.__alloc_page(index)

  def alloc_pages(self, base = None, count = 1):
    """
    Allocate continuous sequence of anonymous pages.

    :param u24 base: if set, start searching pages from this address.
    :param int count: number of requested pages.
    :returns: list of newly allocated pages.
    :rtype: ``list`` of :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.InvalidResourceError: when there is no available sequence of
      pages.
    """

    self.DEBUG('mc.alloc_pages: base=%s, count=%s', UINT32_FMT(base) if base is not None else '<none>', count)

    if base is not None:
      pages_start = base // PAGE_SIZE
      pages_cnt = self.pages_cnt - pages_start
    else:
      pages_start = 0
      pages_cnt = self.pages_cnt

    self.DEBUG('mc.alloc_pages: page=%s, cnt=%s', pages_start, pages_cnt)

    for i in range(pages_start, pages_start + pages_cnt):
      for j in range(i, i + count):
        if j in self.pages:
          break

      else:
        return [self.__alloc_page(j) for j in range(i, i + count)]

    raise InvalidResourceError('No sequence of free pages available')

  def alloc_page(self, base = None):
    """
    Allocate new anonymous page for usage. The first available index is used.

    :param int base: if set, start searching pages from this address.
    :returns: newly reserved page.
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.InvalidResourceError: when there is no available page.
    """

    self.DEBUG('mc.alloc_page: base=%s', UINT32_FMT(base) if base is not None else '<none>')

    if base is not None:
      pages_start = base // PAGE_SIZE
      pages_cnt = self.pages_cnt - pages_start
    else:
      pages_start = 0
      pages_cnt = self.pages_cnt

    self.DEBUG('mc.alloc_page: page=%s, cnt=%s', pages_start, pages_cnt)

    for i in range(pages_start, pages_start + pages_cnt):
      if i not in self.pages:
        self.DEBUG('mc.alloc_page: page=%s', i)
        return self.__alloc_page(i)

    raise InvalidResourceError('No free page available')

  def register_page(self, pg):
    """
    Install page object for a specific memory page. This method is intended
    for external objects, e.g. device drivers to install their memory page
    objects to handle memory-mapped IO.

    :param ducky.mm.MemoryPage pg: page to be installed
    :returns: installed page
    :rtype: :py:class:`ducky.mm.AnonymousMemoryPage`
    :raises ducky.errors.AccessViolationError: when there is already allocated page
    """

    self.DEBUG('mc.register_page: pg=%s', pg)

    if pg.index in self.pages:
      raise AccessViolationError('Page {} is already allocated'.format(pg.index))

    return self.__set_page(pg)

  def unregister_page(self, pg):
    """
    Remove page object for a specific memory page. This method is intende
    for external objects, e.g. device drivers to remove their memory page objects
    handling memory-mapped IO.

    :param ducky.mm.MemoryPage pg: page to be removed
    :raises ducky.errors.AccessViolationError: when there is no allocated page
    """

    self.DEBUG('mc.unregister_page: pg=%s', pg)

    if pg.index not in self.pages:
      raise AccessViolationError('Page {} is not allocated'.format(pg.index))

    self.__remove_page(pg)

  def free_page(self, page):
    """
    Free memory page when it's no longer needed.

    :param ducky.mm.MemoryPage page: page to be freed.
    """

    self.DEBUG('mc.free_page: page=%i, base=%s', page.index, UINT32_FMT(page.base_address))

    self.__remove_page(page)

  def free_pages(self, page, count = 1):
    """
    Free a continuous sequence of pages when they are no longer needed.

    :param ducky.mm.MemoryPage page: first page in series.
    :param int count: number of pages.
    """

    self.DEBUG('mc.free_pages: page=%i, base=%s, count=%s', page.index, UINT32_FMT(page.base_address), count)

    for i in range(page.index, page.index + count):
      self.free_page(self.pages[i])

  def get_page(self, index):
    """
    Return memory page, specified by its index from the beginning of memory.

    :param int index: index of requested page.
    :rtype: :py:class:`ducky.mm.MemoryPage`
    :raises ducky.errors.AccessViolationError: when requested page is not allocated.
    """

    if index not in self.pages:
      return self.alloc_specific_page(index)
      # raise AccessViolationError('Page {} not allocated yet'.format(index))

    return self.pages[index]

  def get_pages(self, pages_start = 0, pages_cnt = None, ignore_missing = False):
    """
    Return list of memory pages.

    :param int pages_start: index of the first page, 0 by default.
    :param int pages_cnt: number of pages to get, number of all memory pages by default.
    :param bool ignore_missing: if ``True``, ignore missing pages, ``False`` by default.
    :raises ducky.errors.AccessViolationError: when ``ignore_missing == False`` and there's
      a missing page in requested range, this exception is rised.
    :returns: list of pages in area
    :rtype: `list` of :py:class:`ducky.mm.MemoryPage`
    """

    self.DEBUG('mc.pages: pages_start=%s, pages_cnt=%s, ignore_missing=%s', pages_start, pages_cnt, ignore_missing)

    pages_cnt = pages_cnt or self.pages_cnt

    if ignore_missing is True:
      return (self.pages[i] for i in range(pages_start, pages_start + pages_cnt) if i in self.pages)
    else:
      return (self.pages[i] for i in range(pages_start, pages_start + pages_cnt))

  def pages_in_area(self, address = 0, size = None, ignore_missing = False):
    """
    Return list of memory pages.

    :param u24 address: beggining address of the area, by default 0.
    :param u24 size: size of the area, by default the whole memory size.
    :param bool ignore_missing: if ``True``, ignore missing pages, ``False`` by default.
    :raises ducky.errors.AccessViolationError: when ``ignore_missing == False`` and there's
      a missing page in requested range, this exception is rised.
    :returns: list of pages in area
    :rtype: `list` of :py:class:`ducky.mm.MemoryPage`
    """

    self.DEBUG('mc.pages_in_area: address=%s, size=%s', UINT32_FMT(address), size)

    size = size or self.size
    pages_start, pages_cnt = area_to_pages(address, size)

    return self.get_pages(pages_start = pages_start, pages_cnt = pages_cnt, ignore_missing = ignore_missing)

  def boot(self):
    """
    Prepare memory controller for immediate usage by other components.
    """

    self.machine.tenh('mm: %s, %s available', sizeof_fmt(self.size, max_unit = 'Ki'), sizeof_fmt(self.size - len(self.pages) * PAGE_SIZE, max_unit = 'Ki'))

  def halt(self):
    pass

  def read_u8(self, addr):
    self.DEBUG('mc.read_u8: addr=%s', UINT32_FMT(addr))

    return self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u8(addr & (PAGE_SIZE - 1))

  def read_u16(self, addr):
    self.DEBUG('mc.read_u16: addr=%s', UINT32_FMT(addr))

    return self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u16(addr & (PAGE_SIZE - 1))

  def read_u32(self, addr):
    self.DEBUG('mc.read_u32: addr=%s', UINT32_FMT(addr))

    return self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).read_u32(addr & (PAGE_SIZE - 1))

  def write_u8(self, addr, value):
    self.DEBUG('mc.write_u8: addr=%s, value=%s', UINT32_FMT(addr), UINT8_FMT(value))

    self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u8(addr & (PAGE_SIZE - 1), value)

  def write_u16(self, addr, value):
    self.DEBUG('mc.write_u16: addr=%s, value=%s', UINT32_FMT(addr), UINT16_FMT(value))

    self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u16(addr & (PAGE_SIZE - 1), value)

  def write_u32(self, addr, value):
    self.DEBUG('mc.write_u32: addr=%s, value=%s', UINT32_FMT(addr), UINT32_FMT(value))

    self.get_page((addr & PAGE_MASK) >> PAGE_SHIFT).write_u32(addr & (PAGE_SIZE - 1), value)



import six.moves
import urllib

if not hasattr(urllib, 'parse'):
  urllib.parse = six.moves.urllib.parse

from .. import patch  # noqa
from ..machine import Machine
from ..util import str2int, UINT32_FMT
from ..streams import OutputStream, InputStream
from ..interfaces import IReactorTask
from ..profiler import STORE

import optparse
import os
import signal
import sys
import threading

from autobahn.twisted.websocket import WebSocketServerProtocol, WebSocketServerFactory
from twisted.internet import reactor

class WSOutputStream(OutputStream):
  """
  Websocket output stream, receiving bytes from TTY frontend, and pushing them
  to protocol's socket.

  :param DuckyProtocol protocol: protocol instance with opened websocket.
  """

  def __init__(self, protocol, *args, **kwargs):
    super(WSOutputStream, self).__init__(*args, **kwargs)

    self._protocol = protocol

  def write(self, buff):
    """
    Write buffer into the socket.

    Called by device from machine thread, therefore this method hands buffer
    over to the reactor thread.

    :param bytearray buff: bytes to send to client.
    """

    self.DEBUG('%s.write: buff=%s', self.__class__.__name__, buff)

    reactor.callFromThread(self._protocol.sendMessage, (''.join([chr(b) for b in buff])).encode('utf8'), isBinary = False)

class WSInputStream(InputStream):
  """
  Websocket input stream, receiving bytes from opened websocket, and pushing
  them to keyboard frontend device.

  Stream has an internal LIFO buffer that is being fed by protocol, and
  consumed by VM frontend device.

  :param DuckyProtocol protocol: protocol instance with opened websocket.
  """

  def __init__(self, protocol, *args, **kwargs):
    super(WSInputStream, self).__init__(*args, **kwargs)

    self._protocol = protocol
    self._buffer = []

    self._flush_task = None

  def has_poll_support(self):
    """
    See :py:meth:`ducky.streams.Stream.has_poll_support'.

    """

    return True

  def register_with_reactor(self, reactor, on_read = None, on_error = None):
    """
    See :py:meth:`ducky.streams.Stream.register_with_reactor'.

    """

    input_buffer = self._buffer

    class FlushTask(IReactorTask):
      def run(self):
        if not input_buffer:
          return

        on_read()

    self._flush_task = FlushTask()
    reactor.add_task(self._flush_task)
    reactor.task_runnable(self._flush_task)

  def unregister_with_reactor(self, reactor):
    """
    See :py:meth:`ducky.streams.Stream.unregister_with_reactor'.

    """

    reactor.remove_task(self._flush_task)

  def enqueue(self, buff):
    """
    Called by protocol instance, to add newly received data to stream's
    buffer.

    :param bytearray buff: recerived bytes.
    """

    for c in buff:
      self._buffer.append(ord(c))

  def read(self, size = None):
    """
    See :py:meth:`ducky.streams.Stream.read'.

    """

    self.DEBUG('%s.read: size=%s', self.__class__.__name__, size)

    ret = []
    if size is None:
      while self._buffer:
        ret.append(self._buffer.pop(0))

    else:
      i = 0
      while self._buffer and i < size:
        ret.append(self._buffer.pop(0))
        i += 1

    if not ret:
      return None

    return ret

def print_machine_stats(logger, M):
  table_exits = [
    ['Core', 'Exit code']
  ]

  table_inst_caches = [
    ['Core', 'Reads', 'Inserts', 'Hits', 'Misses', 'Prunes']
  ]
  table_cnts = [
    ['Core', 'Ticks']
  ]

  def __check_stats(core):
    table_exits.append([str(core), UINT32_FMT(core.exit_code)])

    table_inst_caches.append([
      str(core),
      core.mmu.instruction_cache.reads,
      core.mmu.instruction_cache.inserts,
      core.mmu.instruction_cache.hits,
      core.mmu.instruction_cache.misses,
      core.mmu.instruction_cache.prunes
    ])

    table_cnts.append([
      str(core),
      core.registers.cnt.value
    ])

  for core in M.cores:
    __check_stats(core)

  logger.info('')
  logger.info('Exit codes')
  logger.table(table_exits)
  logger.info('')
  logger.info('Instruction caches')
  logger.table(table_inst_caches)
  logger.info('')
  logger.table(table_cnts)
  logger.info('')

  inst_executed = sum([core.registers.cnt.value for core in M.cores])
  runtime = float(M.end_time - M.start_time)
  if runtime > 0:
    logger.info('Executed instructions: %i %f (%.4f/sec)', inst_executed, runtime, float(inst_executed) / runtime)
  logger.info('')

class DuckyProtocol(WebSocketServerProtocol):
  """
  Protocol handling communication between VM and remote terminal emulator.

  :param logging.Logger logger: ``Logger`` instanceto use for logging.
  :param options: command-line options, as returned by option parser.
  :param ducky.config.MachineConfig config: VM configuration.
  """

  _machines = []

  def __init__(self, logger, options, config):
    super(DuckyProtocol, self).__init__()

    self._logger = logger
    self._options = options
    self._config = config

    self.DEBUG = logger.debug

    self._machine = Machine()
    self._machine_thread = None
    self._input_stream = None

    self._profiler = None

  def onMessage(self, payload, isBinary):
    """
    Called when new data arrived from client. Feeds the data to VM's
    input stream.

    See :py:meth:`autobahn.twisted.websocket.WebSocketServerProtocol.onMessage`.
    """

    self.DEBUG('%s.onMessage', self.__class__.__name__)

    try:
      if self._input_stream is None:
        return

      self._input_stream.enqueue(payload)

    except Exception as e:
      self._logger.exception(e)
      self.sendClose()

  def run_machine(self):
    """
    Wraps VM's ``run()`` method by ``try/except`` clause, and call protocols
    ``sendClose()`` method when VM finishes.

    This is the target of VM`s thread.
    """

    try:
      if self._options.profile:
        self._profiler = STORE.get_machine_profiler()
        self._profiler.enable()

      self._machine.run()

    except Exception as e:
      self._logger.exception(e)

    finally:
      if self._profiler is not None:
        self._profiler.disable()

      reactor.callFromThread(self.sendClose)

  def onOpen(self, *args, **kwargs):
    """
    Called when new client connects to the server.

    This callback will setup and start new VM, connecting it to remote terminal
    by wrapping this protocol instance in two streams (input/output), and spawn
    a fresh new thread for it.
    """

    try:
      self.DEBUG('%s.onOpen', self.__class__.__name__)

      self._machine.hw_setup(self._config)

      self._output_stream = WSOutputStream(self, self._machine, '<ws-out>', fd = 0, close = False, allow_close = False)
      self._input_stream = WSInputStream(self, self._machine, '<ws-in>', fd = 0, close = False, allow_close = False)
      self._machine.get_device_by_name('device-3', klass = 'terminal').enqueue_streams(streams_in = [self._input_stream], stream_out = self._output_stream)

      self._machine.boot()

      for poke in self._options.poke:
        address, value, length = poke.split(':')

        if length not in ('1', '2', '4'):
          raise ValueError('Unknown poke size: poke=%s' % poke)

        self._machine.poke(str2int(address), str2int(value), str2int(length))

      self._machines.append(self._machine)

      self._machine_thread = threading.Thread(target = self.run_machine)
      self._machine_thread.start()

    except Exception as e:
      self._logger.exception(e)
      self.sendClose()

  def onClose(self, wasClean, code, reason):
    """
    Called when connection ends. Tell VM to halt, and wait for its thread
    to finish. Then, print some VM stats.
    """

    self.DEBUG('%s.onClose', self.__class__.__name__)

    if self._machine.halted is not True:
      from ..reactor import CallInReactorTask
      self._machine.reactor.add_event(CallInReactorTask(self._machine.halt))

      if self._machine_thread is not None:
        self._machine_thread.join()

    print_machine_stats(self._logger, self._machine)

    self._machines.remove(self._machine)

class DuckySocketServerFactory(WebSocketServerFactory):
  def __init__(self, logger, options, config, *args, **kwargs):
    super(DuckySocketServerFactory, self).__init__(*args, **kwargs)

    self._logger = logger
    self._options = options
    self._config = config

  def buildProtocol(self, *args, **kwargs):
    proto = DuckyProtocol(self._logger, self._options, self._config)
    proto.factory = self
    return proto

def process_config_options(logger, options, config_file = None, set_options = None, add_options = None, enable_devices = None, disable_devices = None):
  """
  Load VM config file, and apply all necessary changes, as requested by command-line options.

  :param logging.Logger logger: ``Logger`` instance to use for logging.
  :param options: command-line options, as returned by option parser.
  :param string config_file: path to configuration file.
  :rtype: ducky.config.MachineConfig
  :returns: VM configuration.
  """

  logger.debug('process_config_options: config_file=%s, set_options=%s, add_options=%s, enable_devices=%s, disable_devices=%s', config_file, set_options, add_options, enable_devices, disable_devices)

  set_options = set_options or []
  add_options = add_options or []
  enable_devices = enable_devices or []
  disable_devices = disable_devices or []

  from ..config import MachineConfig
  config = MachineConfig()

  if config_file is not None:
    config.read(config_file)

  for section, option, value in set_options:
    if not config.has_section(section):
      logger.error('Unknown config section %s', section)
      continue

    config.set(section, option, value)

  for section, option, value in add_options:
    if not config.has_section(section):
      logger.error('Unknown config section %s', section)
      continue

    if not config.has_option(section, option):
      config.set(section, option, value)

    else:
      config.set(section, option, config.get(section, option) + ', ' + value)

  for dev in enable_devices:
    config.set(dev, 'enabled', True)

  for dev in disable_devices:
    config.set(dev, 'enabled', False)

  if 'JIT' in os.environ:
    config.set('machine', 'jit', os.environ['JIT'] == 'yes')

  else:
    config.set('machine', 'jit', options.jit is True)

  return config

def main():
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  # Machine configuration
  opt_group = optparse.OptionGroup(parser, 'Machine hardware')
  parser.add_option_group(opt_group)
  opt_group.add_option('--machine-config',  dest = 'machine_config',  action = 'store',      default = None,  help = 'Path to machine configuration file')
  opt_group.add_option('--machine-profile', dest = 'machine_profile', action = 'store_true', default = False, help = 'Enable profiling of running binaries')
  opt_group.add_option('--set-option',      dest = 'set_options',     action = 'append',     default = [],    metavar = 'SECTION:OPTION=VALUE', help = 'Set option')
  opt_group.add_option('--add-option',      dest = 'add_options',     action = 'append',     default = [],    metavar = 'SECTION:OPTION=VALUE')
  opt_group.add_option('--enable-device',   dest = 'enable_devices',  action = 'append',     default = [],    metavar = 'DEVICE', help = 'Enable device')
  opt_group.add_option('--disable-device',  dest = 'disable_devices', action = 'append',     default = [],    metavar = 'DEVICE', help = 'Disable device')
  opt_group.add_option('--poke',            dest = 'poke',            action = 'append',     default = [],    metavar = 'ADDRESS:VALUE:<124>', help = 'Modify content of memory before running binaries')
  opt_group.add_option('--jit',             dest = 'jit',             action = 'store_true', default = False, help = 'Optimize instructions')

  # Network options
  opt_group = optparse.OptionGroup(parser, 'Network options')
  parser.add_option_group(opt_group)
  opt_group.add_option('--network', dest = 'network', action = 'store_true', default = False, help = 'Start network service')
  opt_group.add_option('--host', dest = 'host', action = 'store', default = '', metavar = 'HOST', help = 'Listen at HOST address')
  opt_group.add_option('--port', dest = 'port', action = 'store', type = 'int', default = 19000, metavar = 'PORT', help = 'Listen at PORT port')
  opt_group.add_option('--queue', dest = 'queue', action = 'store', type = 'int', default = 10, metavar = 'LENGTH', help = 'Listen queue is LENGTH at max')

  # Debugging options
  opt_group = optparse.OptionGroup(parser, 'Debugging options')
  parser.add_option_group(opt_group)
  opt_group.add_option('--profile', dest = 'profile', action = 'store_true', default = False, help = 'Enable profiler')
  opt_group.add_option('--profile-dir', dest = 'profile_dir', action = 'store', default = None, metavar = 'DIR', help = 'Save profiling data into DIR')

  options, logger = parse_options(parser)

  if options.machine_config is None:
    parser.print_help()
    sys.exit(1)

  if options.profile and options.profile_dir is None:
    parser.print_help()
    sys.exit(1)

  config = process_config_options(logger,
                                  options,
                                  config_file = options.machine_config,
                                  set_options = [(section,) + tuple(option.split('=')) for section, option in (option.split(':') for option in options.set_options)],
                                  add_options = [(section,) + tuple(option.split('=')) for section, option in (option.split(':') for option in options.add_options)],
                                  enable_devices = options.enable_devices,
                                  disable_devices = options.disable_devices)

  if options.profile:
    STORE.enable_machine()

  main_thread_profiler = STORE.get_machine_profiler()
  main_thread_profiler.enable()

  if options.network is True:
    from twisted.python import log
    log.startLogging(sys.stdout)

    factory = DuckySocketServerFactory(logger, options, config)
    factory.protocol = DuckyProtocol

    reactor.listenTCP(options.port, factory, backlog = options.queue, interface = options.host)
    reactor.run()

    exit_code = 0

  else:
    M = Machine()
    M.hw_setup(config)

    def signal_handler(sig, frame):
      if sig == signal.SIGUSR1:
        M.tenh('VM suspended by user')
        M.reactor.add_call(M.suspend)

      elif sig == signal.SIGUSR2:
        M.tenh('VM unsuspended by user')
        M.reactor.add_call(M.wake_up)

      elif sig == signal.SIGINT:
        M.tenh('VM halted by user')
        M.reactor.add_call(M.halt)

      elif sig == signal.SIGSEGV:
        M.tenh('VM snapshot requested')
        M.reactor.add_call(M.snapshot('ducky-snapshot-user.bin'))

    signal.signal(signal.SIGINT,  signal_handler)
    signal.signal(signal.SIGUSR1, signal_handler)
    signal.signal(signal.SIGUSR2, signal_handler)
    signal.signal(signal.SIGSEGV, signal_handler)

    M.boot()

    for poke in options.poke:
      address, value, length = poke.split(':')

      if length not in ('1', '2', '4'):
        raise ValueError('Unkn poke size: poke=%s' % poke)

      M.poke(str2int(address), str2int(value), str2int(length))

    M.run()

    print_machine_stats(logger, M)
    exit_code = 1 if M.exit_code != 0 else 0

  main_thread_profiler.disable()

  if options.profile:
    logger.info('Saving profiling data into %s' % options.profile_dir)
    STORE.save(logger, options.profile_dir)

  sys.exit(exit_code)



import time
import datetime

from six.moves import range

from . import Device, IRQProvider, IOProvider, IRQList
from ..errors import InvalidResourceError
from ..mm import u8_t, UINT16_FMT
from ..reactor import RunInIntervalTask
from ..util import F

DEFAULT_IRQ  = 0x00
DEFAULT_FREQ = 100
DEFAULT_PORT_RANGE = 0x300
PORT_RANGE = 0x0007

class RTCTask(RunInIntervalTask):
  def __init__(self, machine, rtc):
    super(RTCTask, self).__init__(10, self.on_tick)

    self.machine = machine
    self.rtc = rtc

    self.stamp = 0
    self.tick = 0

    self.update_tick()

  def update_tick(self):
    self.tick = 1.0 / float(self.rtc.frequency)
    self.machine.DEBUG('rtc: new frequency: %i => %f' % (self.rtc.frequency, self.tick))

  def on_tick(self, task):
    stamp = time.time()
    diff = stamp - self.stamp
    self.machine.DEBUG('rtc: tick: stamp=%s, last=%s, diff=%s, tick=%s, ?=%s' % (stamp, self.stamp, diff, self.tick, diff < self.tick))
    if diff < self.tick:
      return

    self.machine.DEBUG('rtc: trigger irq')

    self.stamp = stamp

    self.machine.trigger_irq(self.rtc)

class RTC(IRQProvider, IOProvider, Device):
  def __init__(self, machine, name, frequency = None, port = None, irq = None, *args, **kwargs):
    super(RTC, self).__init__(machine, 'rtc', name, *args, **kwargs)

    self.frequency = frequency or DEFAULT_FREQ
    self.port = port or DEFAULT_PORT_RANGE
    self.ports = list(range(port, port + PORT_RANGE))
    self.irq = irq or DEFAULT_IRQ
    self.timer_task = RTCTask(machine, self)

    if self.frequency >= 256:
      raise InvalidResourceError('Maximum RTC ticks per second is 255')

  @staticmethod
  def create_from_config(machine, config, section):
    return RTC(machine,
               section,
               frequency = config.getint(section, 'frequency', DEFAULT_FREQ),
               port = config.getint(section, 'port', DEFAULT_PORT_RANGE),
               irq = config.getint(section, 'irq', IRQList.TIMER))

  def boot(self):
    self.machine.DEBUG('RTC.boot')

    for port in self.ports:
      self.machine.register_port(port, self)

    self.machine.reactor.add_task(self.timer_task)
    self.machine.reactor.task_runnable(self.timer_task)

    now = datetime.datetime.now()

    self.machine.tenh('RTC: time %02i:%02i:%02i, date: %02i/%02i/%02i', now.hour, now.minute, now.second, now.day, now.month, now.year - 2000)

  def halt(self):
    for port in self.ports:
      self.machine.unregister_port(port)

    self.machine.reactor.remove_task(self.timer_task)

  def read_u8(self, port):
    if port not in self.ports:
      raise InvalidResourceError('Unhandled port: %s' % UINT16_FMT(port))

    port -= self.port

    if port == 0x0000:
      return u8_t(self.frequency).value

    now = datetime.datetime.now()

    if port == 0x0001:
      return u8_t(now.second).value

    if port == 0x0002:
      return u8_t(now.minute).value

    if port == 0x0003:
      return u8_t(now.hour).value

    if port == 0x0004:
      return u8_t(now.day).value

    if port == 0x0005:
      return u8_t(now.month).value

    if port == 0x0006:
      return u8_t(now.year - 2000).value

  def write_u8(self, port, value):
    if port not in self.ports:
      raise InvalidResourceError(F('Unhandled port: {port:S}', port = port))

    if port != self.ports[0]:
      raise InvalidResourceError(F('Unable to write to read-only port: port={port:S}, value={value:B}', port = port, value = value))

    self.frequency = value
    self.timer_task.update_tick()



import enum
import importlib

from ..interfaces import IMachineWorker
from ..mm import UINT16_FMT
from ..errors import InvalidResourceError

VIRTUAL_INTERRUPTS = {}

class IRQList(enum.IntEnum):
  """
  List of known IRQ sources.
  """

  # HW devices
  TIMER    = 0
  KEYBOARD = 1
  BIO      = 2

  # SW interrupts and exceptions
  HALT    = 16
  BLOCKIO = 17
  VMDEBUG = 18

  IRQ_COUNT = 32


class IOPorts(enum.IntEnum):
  PORT_COUNT = 65536


class Device(IMachineWorker):
  def __init__(self, machine, klass, name):
    super(Device, self).__init__()

    self.machine = machine
    self.klass = klass
    self.name = name

    self.logger = machine.LOGGER
    self.master = None

  @staticmethod
  def create_from_config(machine, config, section):
    raise NotImplementedError()

  def boot(self):
    self.logger.debug('%s.boot', self.__class__.__name__)
    pass

  def halt(self):
    self.logger.debug('%s.halt', self.__class__.__name__)
    pass

  def is_slave(self):
    return self.master is not None

  def get_master(self):
    return self.master


class DeviceFrontend(Device):
  def set_backend(self, device):
    self.backend = device

class DeviceBackend(Device):
  def set_frontend(self, device):
    self.frontend = device

class IRQProvider(object):
  pass


class IOProvider(object):
  def is_port_protected(self, port, read = True):
    return False

  def read_u8(self, port):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

  def read_u16(self, port):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

  def read_u32(self, port):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

  def write_u8(self, port, value):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

  def write_u16(self, port, value):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

  def write_u32(self, port, value):
    raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))


def get_driver_creator(driver_class):
  driver = driver_class.split('.')

  driver_module = importlib.import_module('.'.join(driver[0:-1]))
  driver_class = getattr(driver_module, driver[-1])

  return driver_class.create_from_config



from pycparser import c_ast

from . import ASTOptVisitor, dump_node

class ConstantFoldingVisitor(ASTOptVisitor):
  priority = 100

  def visit_BinaryOp(self, node):
    self.generic_visit(node)

    L = node.left
    R = node.right

    #
    # If both left and right branches are constants, they can be merged into one
    #
    if isinstance(node.left, c_ast.Constant) and isinstance(node.right, c_ast.Constant):
      self.DEBUG(self.log_prefix + 'Fold: binary op on two constants: %s %s %s', dump_node(self.node_parent), dump_node(node.left), dump_node(node.right))
      self.DOWN()

      if L.type == 'int' and R.type == 'int':
        self.DEBUG(self.log_prefix + 'Fold: both branches are integers, fold into one')

        if node.op == '+':
          i = int(L.value) + int(R.value)

        elif node.op == '*':
          i = int(L.value) * int(R.value)

        else:
          self.WARN(self.log_prefix + '%s: Unhandled binary operation: op=%s', self.__class__.__name__, node.op)

        self.replace_child(node, c_ast.Constant('int', i, coord = node.coord))

        self.UP()

    #
    # Anything multiplied by zero is zero - replace such node with constant
    #
    if node.op == '*' and ((isinstance(L, c_ast.Constant) and L.type == 'int' and L.value == '0') or (isinstance(R, c_ast.Constant) and R.type == 'int' and R.value == '0')):
      self.DEBUG(self.log_prefix + 'Fold: multiply by zero: %s %s %s', dump_node(self.node_parent), dump_node(L), dump_node(R))
      self.DOWN()

      self.replace_child(node, c_ast.Constant('int', 0, coord = node.coord))

      self.UP()

    #
    # Anything multiplied by one keeps its value - discard one branch and replace node with the other
    #
    if node.op == '*' and ((isinstance(L, c_ast.Constant) and L.type == 'int' and L.value == '1') or (isinstance(R, c_ast.Constant) and R.type == 'int' and R.value == '1')):
      self.DEBUG(self.log_prefix + 'Fold: multiply by one: %s %s %s', dump_node(self.node_parent), dump_node(L), dump_node(R))
      self.DOWN()

      keep_node = R if isinstance(L, c_ast.Constant) else L

      self.replace_child(node, keep_node)

      self.UP()

from . import AST_PASSES
AST_PASSES['ast-const-prop'] = ConstantFoldingVisitor



import functools
import sys

from six import iterkeys, itervalues, iteritems
from six.moves import range

from functools import partial

from . import registers
from .. import profiler

from ..interfaces import IMachineWorker, ISnapshotable
from ..mm import UINT8_FMT, UINT16_FMT, UINT32_FMT, PAGE_SIZE, PAGE_MASK, PAGE_SHIFT, PageTableEntry, UINT64_FMT
from .registers import Registers, REGISTER_NAMES, FLAGS
from .instructions import DuckyInstructionSet
from ..errors import AccessViolationError, InvalidResourceError
from ..util import LRUCache, Flags
from ..snapshot import SnapshotNode
from ..devices import IRQList

#: Default IVT address
DEFAULT_IVT_ADDRESS = 0x00000000

#: Default PT address
DEFAULT_PT_ADDRESS = 0x00010000

#: Default size of core instruction cache, in instructions.
DEFAULT_CORE_INST_CACHE_SIZE = 256

class CPUState(SnapshotNode):
  def get_core_states(self):
    return [__state for __name, __state in iteritems(self.get_children()) if __name.startswith('core')]

  def get_core_state_by_id(self, coreid):
    return self.get_children()['core{}'.format(coreid)]

class CPUCoreState(SnapshotNode):
  def __init__(self):
    super(CPUCoreState, self).__init__('cpuid', 'coreid', 'registers', 'exit_code', 'alive', 'running', 'idle', 'ivt_address', 'pt_address', 'pt_enabled', 'flags')

class InterruptVector(object):
  """
  Interrupt vector table entry.
  """

  SIZE = 8

  def __init__(self, ip = 0x0000, sp = 0x0000):
    self.ip = ip
    self.sp = sp

  def __repr__(self):
    return '<InterruptVector: ip=%s, sp=%s>' % (UINT32_FMT(self.ip), UINT32_FMT(self.sp))

class CPUException(Exception):
  """
  Base class for CPU-related exceptions.

  :param string msg: message describing exceptional state.
  :param ducky.cpu.CPUCore core: CPU core that raised exception, if any.
  :param u32_t ip: address of an instruction that caused exception, if any.
  """

  def __init__(self, msg, core = None, ip = None):
    super(CPUException, self).__init__(msg)

    self.core = core
    self.ip = ip

class InvalidOpcodeError(CPUException):
  """
  Raised when unknown or invalid opcode is found in instruction.

  :param int opcode: wrong opcode.
  """

  def __init__(self, opcode, *args, **kwargs):
    super(InvalidOpcodeError, self).__init__('Invalid opcode: opcode={}'.format(opcode), *args, **kwargs)

    self.opcode = opcode

class InvalidInstructionSetError(CPUException):
  """
  Raised when switch to unknown or invalid instruction set is requested.

  :param int inst_set: instruction set id.
  """

  def __init__(self, inst_set, *args, **kwargs):
    super(InvalidInstructionSetError, self).__init__('Invalid instruction set requested: inst_set={}'.format(inst_set), *args, **kwargs)

    self.inst_set = inst_set

def do_log_cpu_core_state(core, logger = None, disassemble = True, inst_set = None):
  """
  Log state of a CPU core. Content of its registers, and other interesting or
  useful internal variables are logged.

  :param ducky.cpu.CPUCore core: core whose state should be logged.
  :param logger: called for each line of output to actualy log it. By default,
    core's :py:meth:`ducky.cpu.CPUCore.DEBUG` method is used.
  """

  logger = logger or core.DEBUG
  inst_set = inst_set or core.instruction_set

  for i in range(0, Registers.REGISTER_SPECIAL, 4):
    regs = [(i + j) for j in range(0, 4) if (i + j) < Registers.REGISTER_SPECIAL]
    s = ['r{:02d}={}'.format(reg, UINT32_FMT(core.registers.map[reg].value)) for reg in regs]
    logger(' '.join(s))

  logger('fp=%s    sp=%s    ip=%s', UINT32_FMT(core.registers.fp.value), UINT32_FMT(core.registers.sp.value), UINT32_FMT(core.registers.ip.value))
  logger('flags=%s', core.flags.to_string())
  logger('cnt=%s, alive=%s, running=%s, idle=%s, exit=%i', core.registers.cnt.value, core.alive, core.running, core.idle, core.exit_code)

  if hasattr(core, 'math_coprocessor'):
    for index, v in enumerate(core.math_coprocessor.registers.stack):
      logger('MC: %02i: %s', index, UINT64_FMT(v.value))

  if hasattr(core, 'control_coprocessor'):
    cp = core.control_coprocessor
    logger('CC: cr0=%s, cr1=%s, cr2=%s, cr3=%s', UINT32_FMT(cp.read_cr0()), UINT32_FMT(cp.read_cr1()), UINT32_FMT(cp.read_cr2()), UINT32_FMT(cp.read_cr3()))

  if disassemble is True:
    if core.current_instruction is not None:
      inst = inst_set.disassemble_instruction(core.LOGGER, core.current_instruction)
      logger('inst-set=%02d current=%s', inst_set.instruction_set_id, inst)
    else:
      logger('inst-set=%02d current=<none>', inst_set.instruction_set_id)
  else:
    logger('inst-set=%02d current=<unknown>', inst_set.instruction_set_id)

  for index, frame in enumerate(core.backtrace()):
    logger('Frame #%i: %s', index, frame)

def log_cpu_core_state(*args, **kwargs):
  """
  This is a wrapper for ducky.cpu.do_log_cpu_core_state function. Its main
  purpose is to be removed when debug mode is not set, therefore all debug
  calls of ducky.cpu.do_log_cpu_core_state will disappear from code,
  making such code effectively "quiet".
  """

  do_log_cpu_core_state(*args, **kwargs)

class StackFrame(object):
  def __init__(self, fp):
    super(StackFrame, self).__init__()

    self.FP = fp
    self.IP = None

  def __getattribute__(self, name):
    if name == 'address':
      return self.FP

    return super(StackFrame, self).__getattribute__(name)

  def __repr__(self):
    return '<StackFrame: FP={}, IP={}>'.format(UINT32_FMT(self.FP), UINT32_FMT(self.IP if self.IP is not None else 0))


class CoreFlags(Flags):
  _flags = ['privileged', 'hwint_allowed', 'equal', 'zero', 'overflow', 'sign']
  _labels = 'PHEZOS'


class InstructionCache(LRUCache):
  """
  Simple instruction cache class, based on LRU dictionary, with a limited size.

  :param ducky.cpu.CPUCore core: CPU core that owns this cache.
  :param int size: maximal number of entries this cache can store.
  """

  def __init__(self, mmu, size, *args, **kwargs):
    super(InstructionCache, self).__init__(mmu.core.cpu.machine.LOGGER, size, *args, **kwargs)

    self.mmu = mmu
    self.core = mmu.core

    if mmu.core.cpu.machine.config.getbool('machine', 'jit', False):
      self.get_object = self.get_object_jit

  def get_object(self, addr):
    """
    Read instruction from memory. This method is responsible for the real job of
    fetching instructions and filling the cache.

    :param u24 addr: absolute address to read from
    :return: instruction
    :rtype: ``InstBinaryFormat_Master``
    """

    core = self.core

    inst, desc, opcode = core.instruction_set.decode_instruction(core.LOGGER, core.MEM_IN32(addr, not_execute = False), core = core)
    return inst, opcode, partial(desc.execute, core, inst)

  def get_object_jit(self, addr):
    """
    Read instruction from memory. This method is responsible for the real job of
    fetching instructions and filling the cache.

    :param u24 addr: absolute address to read from
    :return: instruction
    :rtype: ``InstBinaryFormat_Master``
    """

    core = self.core

    inst, desc, opcode = core.instruction_set.decode_instruction(core.LOGGER, core.MEM_IN32(addr, not_execute = False), core = core)

    fn = desc.jit(core, inst)

    if fn is None:
      return inst, opcode, partial(desc.execute, core, inst)

    return inst, opcode, fn

class MMU(ISnapshotable):
  """
  Memory management unit (aka MMU) provides a single point handling all core's memory operations.
  All memory reads and writes must go through this unit, which is then responsible for all
  translations, access control, and caching.

  :param ducky.cpu.CPUCore core: parent core.
  :param ducky.mm.MemoryController memory_controller: memory controller that provides access
    to the main memory.
  """

  def __init__(self, core, memory_controller):
    super(MMU, self).__init__()

    config = core.cpu.machine.config

    self.core = core
    self.memory = memory_controller

    self.force_aligned_access = config.getbool('memory', 'force-aligned-access', default = False)
    self.pt_address = config.getint('cpu', 'pt-address', DEFAULT_PT_ADDRESS)

    self.pt_enabled = False

    self.pte_cache = {}

    self.DEBUG = core.DEBUG

    self.instruction_cache = InstructionCache(self, config.getint('cpu', 'inst-cache', default = DEFAULT_CORE_INST_CACHE_SIZE))

    self.set_access_methods()

  def __debug_wrapper_read(self, reader, *args, **kwargs):
    self.core.debug.pre_memory(args[0], read = True)

    if not self.core.running:
      return

    value = reader(*args, **kwargs)

    self.core.debug.post_memory(args[0], read = True)

    return value

  def __debug_wrapper_write(self, writer, *args, **kwargs):
    self.core.debug.pre_memory(args[0], read = False)

    if not self.core.running:
      return

    writer(*args, **kwargs)

    self.core.debug.post_memory(args[0], read = False)

  def set_access_methods(self):
    """
    Set parent core's memory-access methods to proper shortcuts.
    """

    self.DEBUG('MMU.set_access_methods')

    self.core.MEM_IN8   = self.full_read_u8
    self.core.MEM_IN16  = self.full_read_u16
    self.core.MEM_IN32  = self.full_read_u32
    self.core.MEM_OUT8  = self.full_write_u8
    self.core.MEM_OUT16 = self.full_write_u16
    self.core.MEM_OUT32 = self.full_write_u32

    if self.core.debug is not None:
      self.core.MEM_IN8   = partial(self.__debug_wrapper_read,  self.core.MEM_IN8)
      self.core.MEM_IN16  = partial(self.__debug_wrapper_read,  self.core.MEM_IN16)
      self.core.MEM_IN32  = partial(self.__debug_wrapper_read,  self.core.MEM_IN32)
      self.core.MEM_OUT8  = partial(self.__debug_wrapper_write, self.core.MEM_OUT8)
      self.core.MEM_OUT16 = partial(self.__debug_wrapper_write, self.core.MEM_OUT16)
      self.core.MEM_OUT32 = partial(self.__debug_wrapper_write, self.core.MEM_OUT32)

  def reset(self):
    self.instruction_cache.clear()

    self.pt_enabled = False
    self.pte_cache = {}

  def halt(self):
    pass

  def release_ptes(self):
    self.DEBUG('MMU.release_ptes')

    self.pte_cache = {}

  def get_pte(self, addr):
    """
    Find out PTE for particular physical address. If PTE is not in internal PTE cache, it is
    fetched from PTE table.

    :param u24 addr: memory address.
    """

    pg_index = (addr & PAGE_MASK) >> PAGE_SHIFT

    self.DEBUG('MMU.get_pte: addr=%s, pte-address=%s', UINT32_FMT(addr), UINT32_FMT(self.pt_address + pg_index))

    if pg_index not in self.pte_cache:
      self.pte_cache[pg_index] = pte = PageTableEntry.from_int(self.memory.read_u8(self.pt_address + pg_index))

    else:
      pte = self.pte_cache[pg_index]

    self.DEBUG('  pte=%s (%s)', pte.to_string(), pte.to_int())

    return pte

  def check_access(self, access, addr, align = None):
    """
    Check attempted access against PTE. Be aware that each check can be turned off by configuration file.

    :param access: ``read``, ``write`` or ``execute``.
    :param u24 addr: memory address.
    :param int align: if set, operation is expected to be aligned to this boundary.
    :raises ducky.errors.AccessViolationError: when access is denied.
    """

    self.DEBUG('MMU.check_access: access=%s, addr=%s', access, UINT32_FMT(addr))

    if self.force_aligned_access and align is not None and addr % align:
      raise AccessViolationError('Not allowed to access unaligned memory: access=%s, address=%s, align=%s' % (access, UINT32_FMT(addr), align))

    pg_index = (addr & PAGE_MASK) >> PAGE_SHIFT

    if self.core.privileged or self.pt_enabled is not True:
      return self.memory.get_page(pg_index)

    pte = self.get_pte(addr)

    if getattr(pte, access) == 1:
      return self.memory.get_page(pg_index)

    raise AccessViolationError('Not allowed to access memory: access=%s, address=%s, pte=%s' % (access, UINT32_FMT(addr), pte.to_string()))

  def full_read_u8(self, addr):
    self.DEBUG('MMU.raw_read_u8: addr=%s', UINT32_FMT(addr))

    return self.check_access('read', addr).read_u8(addr & (PAGE_SIZE - 1))

  def full_read_u16(self, addr):
    self.DEBUG('MMU.raw_read_u16: addr=%s', UINT32_FMT(addr))

    return self.check_access('read', addr, align = 2).read_u16(addr & (PAGE_SIZE - 1))

  def full_read_u32(self, addr, not_execute = True):
    self.DEBUG('MMU.raw_read_u32: addr=%s', UINT32_FMT(addr))

    pg = self.check_access('read', addr, align = 4)

    if not_execute is not True:
      pg = self.check_access('execute', addr)

    return pg.read_u32(addr & (PAGE_SIZE - 1))

  def full_write_u8(self, addr, value):
    self.DEBUG('MMU.raw_write_u8: addr=%s, value=%s', UINT32_FMT(addr), UINT8_FMT(value))

    return self.check_access('write', addr).write_u8(addr & (PAGE_SIZE - 1), value)

  def full_write_u16(self, addr, value):
    self.DEBUG('MMU.raw_write_u16: addr=%s, value=%s', UINT32_FMT(addr), UINT16_FMT(value))

    return self.check_access('write', addr).write_u16(addr & (PAGE_SIZE - 1), value)

  def full_write_u32(self, addr, value):
    self.DEBUG('MMU.raw_write_u32: addr=%s, value=%s', UINT32_FMT(addr), UINT32_FMT(value))

    return self.check_access('write', addr).write_u32(addr & (PAGE_SIZE - 1), value)

class CPUCore(ISnapshotable, IMachineWorker):
  """
  This class represents the main workhorse, one of CPU cores. Reads
  instructions, executes them, has registers, caches, handles interrupts,
  ...

  :param int coreid: id of this core. Usually, it's its serial number but it
    has no special meaning.
  :param ducky.cpu.CPU cpu: CPU that owns this core.
  :param ducky.mm.MemoryController memory_controller: use this controller to
    access main memory.
  """

  def __init__(self, coreid, cpu, memory_controller):
    super(CPUCore, self).__init__()

    config = cpu.machine.config

    self.cpuid = '#{}:#{}'.format(cpu.id, coreid)
    self.cpuid_prefix = self.cpuid + ':'

    def __log(logger, *args, **kwargs):
      args = ('%s ' + args[0],) + (self.cpuid_prefix,) + args[1:]
      logger(*args)

    def __log_exception(logger_fn, exc):
      self.cpu.machine.LOGGER.exception('Exception raised in CPU core')
      do_log_cpu_core_state(self, logger = logger_fn, disassemble = False if isinstance(exc, InvalidOpcodeError) else True)

    self.LOGGER = cpu.machine.LOGGER
    self.DEBUG = lambda *args, **kwargs: __log(self.cpu.machine.DEBUG, *args, **kwargs)
    self.INFO  = lambda *args, **kwargs: __log(self.cpu.machine.INFO, *args, **kwargs)
    self.WARN  = lambda *args, **kwargs: __log(self.cpu.machine.WARN, *args, **kwargs)
    self.ERROR = lambda *args, **kwargs: __log(self.cpu.machine.ERROR, *args, **kwargs)
    self.EXCEPTION = functools.partial(__log_exception, self.ERROR)

    self.id = coreid
    self.cpu = cpu

    self.debug = None

    self.mmu = MMU(self, memory_controller)

    self.registers = registers.RegisterSet()

    self.privileged = True
    self.hwint_allowed = False

    self.arith_equal = False
    self.arith_zero = False
    self.arith_overflow = False
    self.arith_sign = False

    self.ivt_address = config.getint('cpu', 'ivt-address', DEFAULT_IVT_ADDRESS)

    self.instruction_set = DuckyInstructionSet
    self.instruction_set_stack = []

    self.current_ip = None
    self.current_instruction = None

    self.alive = False
    self.running = False
    self.idle = False

    self.core_profiler = profiler.STORE.get_core_profiler(self) if profiler.STORE.is_cpu_enabled() else None

    self.exit_code = 0

    self.frames = []
    self.check_frames = cpu.machine.config.getbool('cpu', 'check-frames', default = False)

    self.coprocessors = {}

    if self.cpu.machine.config.getbool('cpu', 'math-coprocessor', False):
      from .coprocessor import math_copro

      self.math_coprocessor = self.coprocessors['math'] = math_copro.MathCoprocessor(self)

    if config.getbool('cpu', 'control-coprocessor', True):
      from .coprocessor import control
      self.control_coprocessor = self.coprocessors['control'] = control.ControlCoprocessor(self)

  def has_coprocessor(self, name):
    return hasattr(self, '{}_coprocessor'.format(name))

  def __repr__(self):
    return '#{}:#{}'.format(self.cpu.id, self.id)

  def save_state(self, parent):
    self.DEBUG('save_state')

    state = parent.add_child('core{}'.format(self.id), CPUCoreState())

    state.cpuid = self.cpu.id
    state.coreid = self.id

    state.flags = self.flags.to_int()

    state.registers = []

    for i, reg in enumerate(REGISTER_NAMES):
      state.registers.append(int(self.registers.map[reg].value))

    state.ivt_address = self.ivt_address
    state.pt_address = self.mmu.pt_address
    state.pt_enabled = self.mmu.pt_enabled

    state.exit_code = self.exit_code
    state.idle = self.idle
    state.alive = self.alive
    state.running = self.running

    if self.has_coprocessor('math'):
      self.math_coprocessor.save_state(state)

  def load_state(self, state):
    self.flags = CoreFlags.from_int(state.flags)

    for i, reg in enumerate(REGISTER_NAMES):
      self.registers.map[reg].value = state.registers[i]

    self.ivt_address = state.ivt_address
    self.mmu.pt_address = state.pt_address
    self.mmu.pt_enabled = state.pt_enabled

    self.exit_code = state.exit_code
    self.idle = state.idle
    self.alive = state.alive
    self.running = state.running

    if self.has_coprocessor('math'):
      self.math_coprocessor.load_state(state.get_children()['math_coprocessor'])

  def init_debug_set(self):
    if self.debug is None:
      from .. import debugging
      self.debug = debugging.DebuggingSet(self)

      self.mmu.set_access_methods()

  def REG(self, reg):
    return self.registers.map[reg]

  def IP(self):
    return self.registers.ip

  def SP(self):
    return self.registers.sp

  def FP(self):
    return self.registers.fp

  def reset(self, new_ip = 0x00000000):
    """
    Reset core's state. All registers are set to zero, all flags are set to zero,
    except ``HWINT`` flag which is set to one, and ``IP`` is set to requested value.

    :param u32_t new_ip: new ``IP`` value, defaults to zero
    """

    self.instruction_set = DuckyInstructionSet
    self.instruction_set_stack = []

    for reg in registers.RESETABLE_REGISTERS:
      self.REG(reg).value = 0

    self.flags = CoreFlags.create(privileged = True)

    self.registers.ip.value = new_ip

    self.mmu.reset()

  def backtrace(self):
    bt = []

    if self.check_frames:
      for frame in self.frames:
        bt.append(repr(frame))

      return bt

    bt = []

    for frame_index, frame in enumerate(self.frames):
      ip = self.mmu.memory.read_u32(frame.address + 4)
      bt.append(ip)

    ip = self.registers.ip.value - 4
    bt.append(ip)

    return bt

  def raw_push(self, val):
    """
    Push value on stack. ``SP`` is decremented by four, and value is written at this new address.

    :param u32 val: value to be pushed
    """

    self.DEBUG("raw_push: sp=%s, value=%s", UINT32_FMT(self.registers.sp.value), UINT32_FMT(val))

    self.registers.sp.value -= 4
    self.MEM_OUT32(self.registers.sp.value, val)

  def raw_pop(self):
    """
    Pop value from stack. 4 byte number is read from address in ``SP``, then ``SP`` is incremented by four.

    :return: popped value
    :rtype: ``u32``
    """

    ret = self.MEM_IN32(self.registers.sp.value)
    self.registers.sp.value += 4
    return ret

  def push(self, *regs):
    for reg_id in regs:
      value = self.flags.to_int() if reg_id == FLAGS else self.registers.map[reg_id].value

      self.DEBUG('push: %s (%s) at %s', reg_id, UINT32_FMT(value), UINT32_FMT(self.registers.sp.value - 4))
      self.raw_push(value)

  def pop(self, *regs):
    for reg_id in regs:
      value = self.raw_pop()

      if reg_id == FLAGS:
        self.flags = CoreFlags.from_int(value)

      else:
        self.registers.map[reg_id].value = value

      self.DEBUG('pop: %s (%s) from %s', reg_id, UINT32_FMT(value), UINT32_FMT(self.registers.sp.value - 4))

  def create_frame(self):
    """
    Create new call stack frame. Push ``IP`` and ``FP`` registers and set ``FP`` value to ``SP``.
    """

    self.DEBUG('create_frame')

    self.push(Registers.IP, Registers.FP)

    self.registers.fp.value = self.registers.sp.value

    if self.check_frames:
      self.frames.append(StackFrame(self.registers.fp.value))

  def destroy_frame(self):
    """
    Destroy current call frame. Pop ``FP`` and ``IP`` from stack, by popping ``FP`` restores previous frame.

    :raises CPUException: if current frame does not match last created frame.
    """

    self.DEBUG('destroy_frame')

    if self.check_frames:
      if self.frames[-1].FP != self.registers.sp.value:
        raise CPUException('Leaving frame with wrong SP: IP={}, saved SP={}, current SP={}'.format(UINT32_FMT(self.registers.ip.value), UINT32_FMT(self.frames[-1].FP), UINT32_FMT(self.registers.sp.value)))

      self.frames.pop()

    self.pop(Registers.FP, Registers.IP)

  def __load_interrupt_vector(self, index):
    self.DEBUG('load_interrupt_vector: ivt=%s, index=%i', UINT32_FMT(self.ivt_address), index)

    desc = InterruptVector()

    vector_address = self.ivt_address + index * InterruptVector.SIZE

    desc.ip = self.MEM_IN32(vector_address)
    desc.sp = self.MEM_IN32(vector_address + 4)

    return desc

  def __enter_interrupt(self, index):
    """
    Prepare CPU for handling interrupt routine. New stack is allocated, content fo registers
    is saved onto this new stack, and new call frame is created on this stack. CPU is switched
    into privileged mode. ``CS`` and ``IP`` are set to values, stored in interrupt descriptor
    table at specified offset.

    :param u24 table_address: address of interrupt descriptor table
    :param int index: interrupt number, its index into IDS
    """

    self.DEBUG('__enter_interrupt: index=%i', index)

    if index >= IRQList.IRQ_COUNT:
      raise InvalidResourceError('Interrupt index out of range: index=%d' % index)

    iv = self.__load_interrupt_vector(index)

    self.DEBUG('__enter_interrupt: desc=%s', iv)

    old_SP = self.registers.sp.value

    self.registers.sp.value = iv.sp

    self.raw_push(old_SP)
    self.push(FLAGS)
    self.create_frame()

    self.privileged = True

    self.registers.ip.value = iv.ip

    if self.check_frames:
      self.frames[-1].IP = iv.ip

    self.instruction_set_stack.append(self.instruction_set)
    self.instruction_set = DuckyInstructionSet

  def exit_interrupt(self):
    """
    Restore CPU state after running a interrupt routine. Call frame is destroyed, registers
    are restored, stack is returned back to memory pool.
    """

    self.DEBUG('exit_interrupt')

    self.destroy_frame()
    self.pop(FLAGS)

    old_SP = self.raw_pop()

    self.registers.sp.value = old_SP

    self.instruction_set = self.instruction_set_stack.pop(0)

  def do_int(self, index):
    """
    Handle software interrupt. Real software interrupts cause CPU state to be saved
    and new stack and register values are prepared by ``__enter_interrupt`` method,
    virtual interrupts are simply triggered without any prior changes of CPU state.

    :param int index: interrupt number
    """

    self.DEBUG('do_int: %s', index)

    if index in self.cpu.machine.virtual_interrupts:
      self.DEBUG('do_int: calling virtual interrupt')

      self.cpu.machine.virtual_interrupts[index].run(self)

      self.DEBUG('do_int: virtual interrupt finished')

    else:
      self.__enter_interrupt(index)

      self.DEBUG('do_int: CPU state prepared to handle interrupt')

  def __do_irq(self, index):
    """
    Handle hardware interrupt. CPU state is saved and prepared for interrupt routine
    by calling ``__enter_interrupt`` method. Receiving of next another interrupts
    is prevented by clearing ``HWINT`` flag, and ``idle`` flag is set to ``False``.
    """

    self.DEBUG('__do_irq: %s', index)

    self.__enter_interrupt(index)
    self.hwint_allowed = False
    self.change_runnable_state(idle = False)

    self.DEBUG('__do_irq: CPU state prepared to handle IRQ')
    log_cpu_core_state(self, inst_set = self.instruction_set_stack[-1])

  def irq(self, index):
    try:
      self.__do_irq(index)

    except (CPUException, ZeroDivisionError, AccessViolationError) as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

  def __get_flags(self):
    return CoreFlags.create(privileged = self.privileged, hwint_allowed = self.hwint_allowed, equal = self.arith_equal, zero = self.arith_zero, overflow = self.arith_overflow, sign = self.arith_sign)

  def __set_flags(self, flags):
    self.privileged = flags.privileged
    self.hwint_allowed = flags.hwint_allowed
    self.arith_equal = flags.equal
    self.arith_zero = flags.zero
    self.arith_overflow = flags.overflow
    self.arith_sign = flags.sign

  flags = property(__get_flags, __set_flags)

  def check_protected_ins(self):
    """
    Raise ``AccessViolationError`` if core is not running in privileged mode.

    This method should be used by instruction handlers that require privileged mode, e.g. protected instructions.

    :raises AccessViolationError: if the core is not in privileged mode
    """

    if not self.privileged:
      raise AccessViolationError('Instruction not allowed in unprivileged mode: inst={}'.format(self.current_instruction))

  def check_protected_port(self, port):
    if port not in self.cpu.machine.ports:
      raise InvalidResourceError('Unhandled port: port={}'.format(UINT16_FMT(port)))

    if self.privileged:
      return

    if self.cpu.machine.ports[port].is_port_protected(port):
      raise AccessViolationError('Access to port not allowed in unprivileged mode: inst={}, port={}'.format(self.current_instruction, port))

  def step(self):
    """
    Perform one "step" - fetch next instruction, increment IP, and execute instruction's code (see inst_* methods)
    """

    self.DEBUG('----- * ----- * ----- * ----- * ----- * ----- * ----- * -----')

    has_debug = self.debug is not None

    if has_debug:
      self.debug.pre_step()

      if not self.running:
        return

    # Read next instruction
    self.DEBUG('"FETCH" phase')

    ip = self.registers.ip
    self.current_ip = ip.value

    self.DEBUG('fetch instruction: ip=%s', UINT32_FMT(ip.value))

    try:
      self.current_instruction, opcode, execute = self.mmu.instruction_cache[ip.value]
      ip.value += 4

      self.DEBUG('"EXECUTE" phase: %s %s', UINT32_FMT(self.current_ip), self.instruction_set.disassemble_instruction(self.LOGGER, self.current_instruction))
      log_cpu_core_state(self)

      execute()

    except (InvalidOpcodeError, AccessViolationError, InvalidResourceError) as e:
      self.die(e)
      return

    cnt = self.registers.cnt
    cnt.value += 1

    self.DEBUG('"SYNC" phase:')
    log_cpu_core_state(self)

    if self.core_profiler is not None:
      self.core_profiler.take_sample()

    if has_debug:
      self.debug.post_step()

  def change_runnable_state(self, alive = None, running = None, idle = None):
    old_state = self.alive and self.running and not self.idle

    if alive is not None:
      self.alive = alive

    if running is not None:
      self.running = running

    if idle is not None:
      self.idle = idle

    new_state = self.alive and self.running and not self.idle

    if old_state != new_state:
      if new_state is True:
        self.cpu.machine.reactor.task_runnable(self)

      else:
        self.cpu.machine.reactor.task_suspended(self)

  def suspend(self):
    self.DEBUG('CPUCore.suspend')

    self.change_runnable_state(running = False)
    self.cpu.machine.events.trigger('on-core-suspend', self)

  def wake_up(self):
    self.DEBUG('CPUCore.wake_up')

    self.change_runnable_state(running = True)
    self.cpu.machine.events.trigger('on-core-running', self)

  def die(self, exc):
    self.DEBUG('CPUCore.die')

    self.exit_code = 1

    self.EXCEPTION(exc)

    self.halt()

  def halt(self):
    self.DEBUG('CPUCore.halt')

    self.cpu.machine.events.trigger('on-core-suspended', self)
    self.cpu.machine.events.trigger('on-core-halted', self)

    self.mmu.halt()

    self.change_runnable_state(alive = False, running = False)

    log_cpu_core_state(self)

    self.cpu.machine.reactor.remove_task(self)

    self.cpu.machine.tenh('%r: CPU core halted', self)

  def run(self):
    try:
      self.step()

    except (CPUException, ZeroDivisionError, AccessViolationError) as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

    except Exception as e:
      e.exc_stack = sys.exc_info()
      self.die(e)

  def boot(self):
    self.DEBUG('CPUCore.boot')

    from ..boot import DEFAULT_BOOTLOADER_ADDRESS

    self.reset(new_ip = DEFAULT_BOOTLOADER_ADDRESS)

    log_cpu_core_state(self)

    self.cpu.machine.reactor.add_task(self)
    self.change_runnable_state(alive = True, running = True)
    self.cpu.machine.events.trigger('on-core-alive', self)
    self.cpu.machine.events.trigger('on-core-running', self)

    if self.core_profiler is not None:
      self.core_profiler.enable()

    self.cpu.machine.tenh('%r: CPU core is up', self)
    if self.mmu.instruction_cache is not None:
      self.cpu.machine.tenh('%r:  %d IC slots', self, self.mmu.instruction_cache.size)
    self.cpu.machine.tenh('%r:  check-frames: %s', self, 'yes' if self.check_frames else 'no')
    if self.coprocessors:
      self.cpu.machine.tenh('%r:  coprocessor: %s', self, ' '.join(sorted(iterkeys(self.coprocessors))))

class CPU(ISnapshotable, IMachineWorker):
  def __init__(self, machine, cpuid, memory_controller, cores = 1):
    super(CPU, self).__init__()

    self.cpuid_prefix = '#{}:'.format(cpuid)

    def __log(logger, *args):
      args = ('%s ' + args[0],) + (self.cpuid_prefix,) + args[1:]
      logger(*args)

    self.DEBUG = lambda *args: __log(self.machine.DEBUG, *args)
    self.INFO  = lambda *args: __log(self.machine.INFO, *args)
    self.WARN  = lambda *args: __log(self.machine.WARN, *args)
    self.ERROR = lambda *args: __log(self.machine.ERROR, *args)
    self.EXCEPTION = lambda *args: __log(self.machine.EXCEPTION, *args)

    self.machine = machine
    self.id = cpuid

    self.cores = []
    self.living_cores = []
    self.halted_cores = []
    self.running_cores = []
    self.suspended_cores = []

    for i in range(0, cores):
      __core = CPUCore(i, self, memory_controller)
      self.cores.append(__core)

      self.halted_cores.append(__core)
      self.suspended_cores.append(__core)

    self.machine.console.register_commands([
      ('sc', cmd_set_core),
      ('st', cmd_core_state),
      ('cont', cmd_cont),
      ('step', cmd_step),
    ])

  def __repr__(self):
    return '#%i' % self.id

  def save_state(self, parent):
    state = parent.add_child('cpu{}'.format(self.id), CPUState())

    for core in self.cores:
      core.save_state(state)

  def load_state(self, state):
    for core_state in itervalues(state.get_children()):
      self.cores[core_state.coreid].load_state(core_state)

  def on_core_alive(self, core):
    """
    Triggered when one of cores goes alive.
    """

    self.DEBUG('%s.on_core_alive: core=%s', self.__class__.__name__, core)

    if core.cpu != self:
      return

    self.halted_cores.remove(core)
    self.living_cores.append(core)

  def on_core_halted(self, core):
    """
    Signal CPU that one of cores is no longer alive.
    """

    self.DEBUG('%s.on_core_halted: core=%s', self.__class__.__name__, core)

    if core.cpu != self:
      return

    self.living_cores.remove(core)
    self.halted_cores.append(core)

  def on_core_running(self, core):
    """
    Signal CPU that one of cores is now running.
    """

    self.DEBUG('%s.on_core_running: core=%s', self.__class__.__name__, core)

    if core.cpu != self:
      return

    self.suspended_cores.remove(core)
    self.running_cores.append(core)

  def on_core_suspended(self, core):
    """
    Signal CPU that one of cores is now suspended.
    """

    self.DEBUG('%s.on_core_suspended: core=%s', self.__class__.__name__, core)

    if core.cpu != self:
      return

    self.running_cores.remove(core)
    self.suspended_cores.append(core)

  def suspend(self):
    self.DEBUG('CPU.suspend')

    for core in self.running_cores:
      core.suspend()

  def wake_up(self):
    self.DEBUG('CPU.wake_up')

    for core in self.suspended_cores:
      core.wake_up()

  def die(self, exc):
    self.DEBUG('CPU.die')

    self.EXCEPTION(exc)

    self.halt()

  def halt(self):
    self.DEBUG('CPU.halt')

    for core in self.living_cores:
      core.halt()

    self.machine.events.remove_listener('on-core-alive', self.on_core_alive)
    self.machine.events.remove_listener('on-core-halted', self.on_core_halted)
    self.machine.events.remove_listener('on-core-running', self.on_core_running)
    self.machine.events.remove_listener('on-core-suspended', self.on_core_suspended)

    self.machine.tenh('%r: CPU halted', self)

  def boot(self):
    self.DEBUG('CPU.boot')

    self.machine.events.add_listener('on-core-alive', self.on_core_alive)
    self.machine.events.add_listener('on-core-halted', self.on_core_halted)
    self.machine.events.add_listener('on-core-running', self.on_core_running)
    self.machine.events.add_listener('on-core-suspended', self.on_core_suspended)

    for core in self.cores:
      core.boot()

    self.machine.tenh('%r: CPU is up', self)

def cmd_set_core(console, cmd):
  """
  Set core address of default core used by control commands: sc <coreid>
  """

  M = console.master.machine

  try:
    core = M.core(cmd[1])

  except InvalidResourceError:
    console.writeln('go away')
    return

  console.default_core = core

  console.writeln('# OK: default core is %s', core.cpuid)

def cmd_cont(console, cmd):
  """
  Continue execution until next breakpoint is reached: cont
  """

  if console.default_core is None:
    console.writeln('# ERR: no core selected')
    return

  if console.default_core.running:
    console.writeln('# ERR: core is not suspended')
    return

  console.default_core.wake_up()

  console.writeln('# OK')

def cmd_step(console, cmd):
  """
  Step one instruction forward
  """

  if console.default_core is None:
    console.writeln('# ERR: no core selected')
    return

  if console.default_core.running:
    console.writeln('# ERR: core is not suspended')
    return

  console.default_core.run()

  console.writeln('# OK')

def cmd_next(console, cmd):
  """
  Proceed to the next instruction in the same stack frame.
  """

  core = console.default_core if hasattr(console, 'default_core') else console.machine.cpus[0].cores[0]

  if not core.is_suspended():
    return

  def __ip_addr(offset = 0):
    return core.registers.ip.value + offset

  try:
    inst = core.instruction_set.decode_instruction(core.LOGGER, core.memory.read_u32(__ip_addr()))

    if inst.opcode == core.instruction_set.opcodes.CALL:
      from ..debugging import add_breakpoint

      add_breakpoint(core, core.registers.ip.value + 4, ephemeral = True)

      core.wake_up()

    else:
      core.step()
      core.check_for_events()

      log_cpu_core_state(core, logger = core.INFO)

  except CPUException as e:
    core.die(e)

def cmd_core_state(console, cmd):
  """
  Print core state
  """

  M = console.master.machine
  core = console.default_core if console.default_core is not None else M.cpus[0].cores[0]

  do_log_cpu_core_state(core, logger = functools.partial(console.log, core.INFO))

def cmd_bt(console, cmd):
  """
  Print current backtrace
  """

  M = console.master.machine
  core = console.default_core if console.default_core is not None else M.cpus[0].cores[0]

  table = [
    ['Index', 'symbol', 'offset', 'ip']
  ]

  for index, (ip, symbol, offset) in enumerate(core.backtrace()):
    table.append([index, symbol, UINT32_FMT(offset), UINT32_FMT(ip)])

  console.table(table)



from . import Device

class SnapshotStorage(Device):
  def __init__(self, machine, name, *args, **kwargs):
    super(SnapshotStorage, self).__init__(machine, 'snapshot', name, *args, **kwargs)

  @staticmethod
  def create_from_config(machine, config, section):
    return SnapshotStorage(machine, section)

  def save_snapshot(self, snapshot):
    pass

  def halt(self):
    self.machine.capture_state()


class FileSnapshotStorage(SnapshotStorage):
  def __init__(self, machine, name, filepath = None, *args, **kwargs):
    super(FileSnapshotStorage, self).__init__(machine, name, *args, **kwargs)

    self.filepath = filepath

  @staticmethod
  def create_from_config(machine, config, section):
    return FileSnapshotStorage(machine, section, filepath = config.get(section, 'filepath', None))

  def save_snapshot(self, snapshot):
    snapshot.save(self.filepath)
    self.machine.tenh('snapshot: saved in file %s', self.filepath)

  def boot(self):
    self.machine.tenh('snapshot: storage ready, backed by file %s', self.filepath)

  def halt(self):
    super(FileSnapshotStorage, self).halt()

    self.save_snapshot(self.machine.last_state)

class DefaultFileSnapshotStorage(FileSnapshotStorage):
  @staticmethod
  def create_from_config(machine, config, section):
    return DefaultFileSnapshotStorage(machine, section, filepath = 'ducky-snapshot.bin')



from six import print_, iteritems
from six.moves import cPickle as pickle

from .util import BinaryFile

class SnapshotNode(object):
  def __init__(self, *fields):
    self.__children = {}
    self.__fields = fields

    for field in fields:
      setattr(self, field, None)

  def add_child(self, name, child):
    self.__children[name] = child
    return child

  def get_child(self, name):
    return self.__children[name]

  def get_children(self):
    return self.__children

  def print_node(self, level = 0):
    offset = '    ' * level

    print_(offset, self.__class__.__name__)

    for field in self.__fields:
      print_(offset, '  ', '{}: {}'.format(field, getattr(self, field)))

    if self.__children:
      print_(offset, '  children:')

      for name, value in iteritems(self.__children):
        if isinstance(value, SnapshotNode):
          print_(offset, '    ', '{}: {}'.format(name, value.__class__.__name__))
          value.print_node(level = level + 2)

        else:
          print_(offset, '    ', '{}: {}'.format(name, value))

class VMState(SnapshotNode):
  def __init__(self, logger):
    super(VMState, self).__init__()

    self.logger = logger

  @staticmethod
  def capture_vm_state(machine, suspend = True):
    machine.DEBUG('capture_vm_state')

    state = VMState(machine.LOGGER)

    if suspend:
      machine.DEBUG('suspend vm...')
      machine.suspend()

    machine.DEBUG('capture state...')
    machine.save_state(state)

    if suspend:
      machine.DEBUG('wake vm up...')
      machine.wake_up()

    return state

  @staticmethod
  def load_vm_state(logger, filename):
    return CoreDumpFile.open(logger, filename, 'r').load()

  def save(self, filename):
    with CoreDumpFile.open(self.logger, filename, 'w') as f_out:
      f_out.save(self)

class CoreDumpFile(BinaryFile):
  @staticmethod
  def open(*args, **kwargs):
    return BinaryFile.do_open(*args, klass = CoreDumpFile, **kwargs)

  def load(self):
    self.DEBUG('CoreDumpFile.load')

    return pickle.load(self)

  def save(self, state):
    self.DEBUG('CoreDumpFile.save: state=%s', state)

    logger, state.logger = state.logger, None
    pickle.dump(state, self)
    state.logger = logger



import collections

from pycparser import c_ast

import ducky.cpu
import ducky.util

from . import dump_node, UndefinedStructMemberError

class CType(object):
  types = {}

  def __init__(self, visitor, decl = None):
    self.visitor = visitor
    self.decl = decl

  def __repr__(self):
    raise NotImplementedError('Type %s does not provide its representation' % self.__class__.__name__)

  def __len__(self):
    raise NotImplementedError('Type %s does not provide its size' % self.__class__.__name__)

  def __cmp__(self, other):
    return 0 if repr(self) == repr(other) else -1

  def __eq__(self, other):
    return repr(self) == repr(other)

  def __neq__(self, other):
    return repr(self) != repr(other)

  @staticmethod
  def get_from_desc(visitor, desc):
    visitor.DEBUG(visitor.log_prefix + 'Type.get_from_desc: desc=%s', desc)
    visitor.DOWN()

    T = visitor.types.get(desc)

    if T is None:
      raise RuntimeError('Unknown type: %s' % desc)

    visitor.UP()
    return T

  @staticmethod
  def get_from_decl(visitor, decl):
    visitor.DEBUG(visitor.log_prefix + 'Type.get_from_decl: decl=%s', dump_node(decl))
    visitor.DOWN()

    T = None

    if isinstance(decl, c_ast.PtrDecl):
      T = PointerType(CType.get_from_decl(visitor, decl.type), visitor, decl)

    elif isinstance(decl.type, c_ast.IdentifierType):
      name = ' '.join(decl.type.names)

    elif isinstance(decl.type, c_ast.Struct):
      name = 'struct ' + decl.type.name

    elif isinstance(decl.type, c_ast.PtrDecl):
      T = PointerType(CType.get_from_decl(visitor, decl.type), visitor, decl)

    elif isinstance(decl, c_ast.ArrayDecl):
      T = ArrayType(CType.get_from_decl(visitor, decl.type), visitor, decl)

    else:
      visitor.WARN(visitor.log_prefix + 'Unhandled get_from_decl branch: %s', decl.type.__class__.__name__)
      visitor.UP()
      return None

    if T is None:
      visitor.DEBUG(visitor.log_prefix + 'searching for "%s"', name)
      visitor.DEBUG(visitor.log_prefix + 'known types: %s', visitor.types)

      T = visitor.types.get(name)

    if T is None:
      raise RuntimeError('Unknown type: %s' % name)

    visitor.UP()
    return T

  @staticmethod
  def create_from_decl(visitor, decl):
    visitor.DEBUG(visitor.log_prefix + 'Type.create_from_decl: decl=%s', dump_node(decl))
    visitor.DOWN()

    T = None

    if isinstance(decl, c_ast.PtrDecl):
      T = PointerType(CType.get_from_decl(visitor, decl.type), visitor, decl)

    elif isinstance(decl, c_ast.FuncDecl):
      T = FunctionType(visitor, decl)

    elif isinstance(decl, c_ast.ArrayDecl):
      T = ArrayType(CType.get_from_decl(visitor, decl.type), visitor, decl)

    elif isinstance(decl.type, c_ast.Struct):
      if decl.type.decls is None and decl.type.name is not None:
        T = CType.get_from_decl(visitor, decl)

      else:
        name = visitor.get_new_label() if decl.type.name is None else decl.type.name
        visitor.DEBUG(visitor.log_prefix + 'Type.create_from_decl: struct: %s', dump_node(decl.type))
        T = StructType(name, visitor, decl.type)

    elif isinstance(decl.type, c_ast.IdentifierType):
      T = CType.get_from_decl(visitor, decl)

    else:
      visitor.WARN(visitor.log_prefix + 'Unhandled branch')

    visitor.DEBUG(visitor.log_prefix + 'Type.create_from_decl: %s => %s', dump_node(decl), T)

    visitor.UP()
    return T

class VoidType(CType):
  def __repr__(self):
    return 'void'

  def __len__(self):
    return 0

class IntType(CType):
  def __repr__(self):
    return 'int'

  def __len__(self):
    return 4

class UnsignedIntType(IntType):
  def __repr__(self):
    return 'unsigned int'

class CharType(CType):
  def __repr__(self):
    return 'char'

  def __len__(self):
    return 1

class UnsignedCharType(CharType):
  def __repr__(self):
    return 'unsigned char'

class PointerType(CType):
  def __init__(self, ptr_to_type, *args, **kwargs):
    super(PointerType, self).__init__(*args, **kwargs)

    self.ptr_to_type = ptr_to_type

  def __repr__(self):
    return repr(self.ptr_to_type) + '*'

  def __len__(self):
    return 4

class StructType(CType):
  def __init__(self, name, *args, **kwargs):
    super(StructType, self).__init__(*args, **kwargs)

    self.name = name
    self.fields = collections.OrderedDict()

    self._size = 0

    class Field(object):
      def __init__(self, name, type, offset):
        self.name = name
        self.type = type
        self.offset = offset

    field_offset = 0
    for field in self.decl.decls:
      if isinstance(field.type, c_ast.PtrDecl):
        field_type = PointerType(CType.get_from_decl(self.visitor, field.type.type), self.visitor, field.type.type)

      else:
        field_type = CType.get_from_decl(self.visitor, field.type)

      self.fields[field.name] = field = Field(field.name, field_type, field_offset)

      field_offset = ducky.util.align(4, field_offset + len(field.type))

    self._size = ducky.util.align(4, field_offset)

  def __repr__(self):
    return 'struct %s' % self.name

  def __len__(self):
    return self._size

  def field_offset(self, name):
    if name not in self.fields:
      raise UndefinedStructMemberError(None, self, name)

    return self.fields[name].offset

  def field_type(self, name):
    return self.fields[name].type

class FunctionType(CType):
  def __init__(self, *args, **kwargs):
    super(FunctionType, self).__init__(*args, **kwargs)

    self.rtype = None
    self.args = []

    self.visitor.DEBUG(self.visitor.log_prefix + 'create rtype')
    self.rtype = CType.get_from_decl(self.visitor, self.decl.type)

    if self.decl.args is not None:
      for arg in self.decl.args.params:
        self.args.append(CType.get_from_decl(self.visitor, arg.type))

  def __repr__(self):
    return '%s(%s)' % (repr(self.rtype), ', '.join([repr(param) for param in self.args]))

class ArrayType(CType):
  def __init__(self, item_type, size = None, *args, **kwargs):
    super(ArrayType, self).__init__(*args, **kwargs)

    self.item_type = item_type
    self.size = None

  def __repr__(self):
    return repr(self.item_type) + '[]'

  def __len__(self):
    if self.size is not None:
      return self.size
    return 4



"""
This module provides simple reactor core that runs each of registered tasks at
least once during one iteration of its internal loop.

There are two different kinds of objects that reactor manages:

- task - it's called periodicaly, at least once in each reactor loop iteration
- event - asynchronous events are queued and executed before running any tasks.
  If there are no runnable tasks, reactor loop waits for incomming events.
"""

import collections
import errno
import select

from .interfaces import IReactorTask

FDCallbacks = collections.namedtuple('FDCallbacks', ['on_read', 'on_write', 'on_error'])

class CallInReactorTask(IReactorTask):
  """
  This task request running particular function during the reactor loop. Useful
  for planning future work, and for running tasks in reactor's thread.

  :param fn: callback to fire.
  :param args: arguments for callback.
  :param kwargs: keyword arguments for callback.
  """

  def __init__(self, fn, *args, **kwargs):
    self.fn = fn
    self.args = args
    self.kwargs = kwargs

  def run(self):
    self.fn(*self.args, **self.kwargs)

class RunInIntervalTask(IReactorTask):
  """
  This task will run its callback every ``ticks`` iterations of reactor's main loop.

  :param int ticks: number of main loop iterations between two callback calls.
  :param args: arguments for callback.
  :param kwargs: keyword arguments for callback.
  """

  def __init__(self, ticks, fn, *args, **kwargs):
    self.ticks = ticks
    self.counter = 0

    self.fn = fn
    self.args = args
    self.kwargs = kwargs

  def run(self):
    self.counter += 1

    if self.counter < self.ticks:
      return

    self.counter = 0

    self.fn(self, *self.args, **self.kwargs)

class SelectTask(IReactorTask):
  """
  Private task, serving as a single point where ``select`` syscall is being
  executed. When a subsystem is interested in IO on a file descriptor, such
  file descriptor should be set as non-blocking, and then registered with
  reactor - it's not viable to place ``select`` calls everywhere in different
  drivers. This task takes list of registered file descriptors, checks for
  possible IO oportunities, and fires callbacks accordingly.

  :param ducky.machine.Machine machine: VM this task (and reactor) belongs to.
  :param dict fds: dictionary, where keys are descriptors, and values are lists
    of their callbacks.
  """

  def __init__(self, machine, fds, *args, **kwargs):
    super(SelectTask, self).__init__(*args, **kwargs)

    self.machine = machine

    self.fds = fds
    self.real_fds = {}

    self.poll = select.poll()

  def add_fd(self, fd, on_read = None, on_write = None, on_error = None):
    """
    Register file descriptor with reactor. File descriptor will be checked for
    read/write/error posibilities, and appropriate callbacks will be fired.

    No arguments are passed to callbacks.

    :param int fd: file descriptor.
    :param on_read: function that will be called when file descriptor is available for
      reading.
    :param on_write: function that will be called when file descriptor is available for
      write.
    :param on_error: function that will be called when error state raised on file
      descriptor.
    """

    self.machine.DEBUG('%s.add_fd: fd=%s, on_read=%s, on_write=%s, on_error=%s', self.__class__.__name__, fd, on_read, on_write, on_error)

    if not isinstance(fd, int):
      fd = fd.fileno()

    self.fds[fd] = FDCallbacks(on_read, on_write, on_error)
    self.poll.register(fd, select.POLLERR | (select.POLLIN if on_read is not None else 0) | (select.POLLOUT if on_write is not None else 0))

  def remove_fd(self, fd):
    """
    Unregister file descriptor. It will no longer be checked by its main loop.

    :param int fd: previously registered file descriptor.
    """

    self.machine.DEBUG('%s.remove_fd: fd=%s', self.__class__.__name__, fd)

    if not isinstance(fd, int):
      fd = fd.fileno()

    self.poll.unregister(fd)
    del self.fds[fd]

  def run(self):
    self.machine.DEBUG('%s.run: fds=%s', self.__class__.__name__, self.fds.keys())

    try:
      events = self.poll.poll(0.1)

    except select.error as e:
      if e.args[0] == errno.EINTR:
        self.machine.DEBUG('%s.run: interrupted syscall', self.__class__.__name__)
        return

      raise e

    if not events:
      return

    self.machine.DEBUG('%s.run: events=%s', self.__class__.__name__, events)

    for fd, events in events:
      callbacks = self.fds[fd]

      if events & select.POLLERR:
        if callbacks.on_error is None:
          self.machine.WARN('  unhandled error: fd=%s', fd)
          continue

        self.machine.DEBUG('  trigger err: fd=%s, handler=%s', fd, callbacks.on_error)
        callbacks.on_error()
        continue

      if events & select.POLLIN and callbacks.on_read is not None:
        self.machine.DEBUG('  trigger read: fd=%s, handler=%s', fd, callbacks.on_read)
        callbacks.on_read()

      if events & select.POLLOUT and callbacks.on_write is not None:
        self.machine.DEBUG('  trigger err: fd=%s, handler=%s', fd, callbacks.on_write)
        callbacks.on_write()

class Reactor(object):
  """
  Main reactor class.
  """

  def __init__(self, machine):
    self.machine = machine

    self.tasks = []
    self.runnable_tasks = []
    self.events = []

    self.fds = {}
    self.fds_task = SelectTask(self.machine, self.fds)

  def add_task(self, task):
    """
    Register task with reactor's main loop.
    """

    self.tasks.append(task)

  def remove_task(self, task):
    """
    Unregister task, it will never be ran again.
    """

    self.task_suspended(task)
    self.tasks.remove(task)

  def task_runnable(self, task):
    """
    If not yet marked as such, task is marked as runnable, and its ``run()``
    method will be called every iteration of reactor's main loop.
    """

    if task not in self.runnable_tasks:
      self.runnable_tasks.append(task)

  def task_suspended(self, task):
    """
    If runnable, task is marked as suspended, not runnable, and it will no
    longer be ran by reactor. It's still registered, so reactor's main loop
    will not quit, and task can be later easily re-enabled by calling
    :py:meth:`ducky.reactor.Reactor.task_runnable`.
    """

    if task in self.runnable_tasks:
      self.runnable_tasks.remove(task)

  def add_event(self, event):
    """
    Enqueue asynchronous event.
    """

    self.events.append(event)

  def add_call(self, fn, *args, **kwargs):
    """
    Enqueue function call. Function will be called in reactor loop.
    """

    self.add_event(CallInReactorTask(fn, *args, **kwargs))

  def add_fd(self, fd, on_read = None, on_write = None, on_error = None):
    """
    Register file descriptor with reactor. File descriptor will be checked for
    read/write/error posibilities, and appropriate callbacks will be fired.

    No arguments are passed to callbacks.

    :param int fd: file descriptor.
    :param on_read: function that will be called when file descriptor is available for
      reading.
    :param on_write: function that will be called when file descriptor is available for
      write.
    :param on_error: function that will be called when error state raised on file
      descriptor.
    """

    self.machine.DEBUG('%s.add_fd: fd=%s, on_read=%s, on_write=%s, on_error=%s', self.__class__.__name__, fd, on_read, on_write, on_error)
    self.fds_task.add_fd(fd, on_read = on_read, on_write = on_write, on_error = on_error)

    if len(self.fds) == 1:
      self.add_task(self.fds_task)
      self.task_runnable(self.fds_task)

  def remove_fd(self, fd):
    """
    Unregister file descriptor. It will no longer be checked by its main loop.

    :param int fd: previously registered file descriptor.
    """

    self.machine.DEBUG('Reactor.remove_fd: fd=%s', fd)
    self.fds_task.remove_fd(fd)

    if not self.fds:
      self.remove_task(self.fds_task)

  def run(self):
    """
    Starts reactor loop. Enters endless loop, calling runnable tasks and events,
    and - in case there are no runnable tasks - waits for new events.

    When there are no tasks managed by reactor, loop quits.
    """

    while True:
      if not self.tasks:
        break

      if self.runnable_tasks:
        for task in self.runnable_tasks:
          task.run()

        while self.events:
          e = self.events.pop(0)
          e.run()

      else:
        # This would be better with some sort of interruptible sleep...
        # Maybe use an Event for that, and avoid using Queue when it's
        # not necessary. But that needs more testing, and since I don't
        # have much use for machine that's totally idle, that will come
        # one day in the future
        import time
        time.sleep(0.01)



import tabulate

class ConsoleConnection(object):
  def __init__(self, cid, master, stream_in, stream_out):
    super(ConsoleConnection, self).__init__()

    self.cid = cid

    self.master = master
    self.stream_in = stream_in
    self.stream_out = stream_out

    self.history = []
    self.history_index = 0

    self.buff = []
    self.history.insert(0, self.buff)

    self.default_core = None

  def write(self, buff, *args):
    if isinstance(buff, list):
      buff = ''.join([chr(c) for c in buff])

    if args:
      buff = buff % args

    self.stream_out.write(buff)
    self.stream_out.flush()

  def writeln(self, buff, *args):
    if isinstance(buff, list):
      buff = ''.join([chr(c) for c in buff])

    self.write(buff + '\n', *args)

  def prompt(self):
    self.write('#> ')

  def table(self, table, **kwargs):
    for line in tabulate.tabulate(table, headers = 'firstrow', tablefmt = 'simple', numalign = 'right', **kwargs).split('\n'):
      self.write(line + '\n')

  def log(self, logger, msg, *args):
    self.writeln(msg, *args)
    logger('[console %i] ' + msg, *((self.cid,) + args))

  def execute(self, cmd):
    if cmd[0] not in self.master.commands:
      self.writeln('Unknown command: cmd="%s"', cmd)
      return

    cmd_desc = self.master.commands[cmd[0]]

    try:
      cmd_desc[0](self, cmd, *cmd_desc[1], **cmd_desc[2])

    except Exception as exc:
      self.master.machine.EXCEPTION(exc)

  def boot(self):
    self.prompt()

    self.master.machine.reactor.add_fd(self.stream_in.get_selectee(), on_read = self.read_input, on_error = self.halt)

  def halt(self):
    self.master.machine.reactor.remove_fd(self.stream_in.fileno())

  def die(self, exc):
    self.master.machine.EXCEPTION(exc)
    self.halt()

  def read_input(self):
    def __clear_line():
      self.write([27, 91, 50, 75, 13])

    def __clear_line_from_cursor():
      self.write([27, 91, 75])

    def __move_backward(count = 1):
      self.write([27, 91, count, 68])

    c = ord(self.stream_in.read(1))

    if c == ord('\n'):
      if self.history_index == 0:
        self.history[0] = ''.join([chr(d) for d in self.buff])

      else:
        self.history.pop(0)
        self.history_index -= 1

      line = self.history[self.history_index]
      self.buff = []

      self.writeln('')

      if not line:
        self.history.pop(0)
        self.history = [self.buff] + self.history
        self.prompt()
        return

      cmd = [e.strip() for e in line.split(' ')]
      self.execute(cmd)
      self.prompt()
      return

    self.buff.append(c)

    if c == 127:
      self.buff[-1:] = []

      if self.buff:
        self.buff[-1:] = []

        __clear_line()
        self.prompt()
        self.write(self.buff)
        return

    if len(self.buff) >= 3:
      # up arrow
      if self.buff[-3:] == [27, 91, 65]:
        if self.history_index < len(self.history) - 1:
          self.history_index += 1

        self.buff[-3:] = []
        __clear_line()
        self.prompt()
        self.write(self.history[self.history_index])
        return

      # down arrow
      if self.buff[-3:] == [27, 91, 66]:
        if self.history_index > 0:
          self.history_index -= 1

        self.buff[-3:] = []
        __clear_line()
        self.prompt()
        self.write(self.history[self.history_index])
        return

    self.write(chr(c))

class TerminalConsoleConnection(ConsoleConnection):
  def __init__(self, cid, master):
    import sys
    import termios
    import tty

    self.old_term_settings = termios.tcgetattr(sys.stdin)
    tty.setcbreak(sys.stdin.fileno())

    super(TerminalConsoleConnection, self).__init__(cid, master, sys.stdin, sys.stdout)

  def halt(self):
    super(TerminalConsoleConnection, self).halt()

    import termios
    termios.tcsetattr(self.stream_in, termios.TCSADRAIN, self.old_term_settings)

class ConsoleMaster(object):
  console_id = 0

  def __init__(self, machine):
    super(ConsoleMaster, self).__init__()

    self.machine = machine

    self.connections = []
    self.commands = {}

  def register_command(self, name, callback, *args, **kwargs):
    self.commands[name] = (callback, args, kwargs)

  def is_registered_command(self, name):
    return name in self.commands

  def unregister_command(self, name):
    if name in self.commands:
      del self.commands[name]

  def register_commands(self, commands, *args, **kwargs):
    for name, handler in commands:
      if self.is_registered_command(name):
        continue

      self.register_command(name, handler, *args, **kwargs)

  def connect(self, slave):
    self.connections.append(slave)

  def boot(self):
    self.register_command('?', cmd_help)

  def halt(self):
    self.unregister_command('?')

    for slave in self.connections:
      slave.halt()

def cmd_help(console, cmd):
  """
  List all available command and their descriptions
  """

  table = [
    ['Command', 'Description']
  ]

  for cmd_name in sorted(console.master.commands.keys()):
    table.append([cmd_name, console.master.commands[cmd_name][0].__doc__])

  console.table(table)



"""
*Terminal* is a device that groups together character two input and output
devices, thus forming a simple channel for bidirectional communication
between VM and user.

Terminal has two slave frontends:
  - *input*, usually a keyboard
  - *output*, from simple TTY to more powerful devices

Terminal then manages input and input streams, passing them to its slave
devices, which then transports events between streams and VM's comm channel.
"""

import os

from . import DeviceFrontend, get_driver_creator
from ..streams import InputStream, OutputStream

def parse_io_streams(machine, config, section):
  streams_in, stream_out = None, None

  if config.has_option(section, 'streams_in'):
    streams_in = [InputStream.create(machine, e.strip()) for e in config.get(section, 'streams_in').split(',')]

  if config.has_option(section, 'stream_out'):
    stream_out = OutputStream.create(machine, config.get(section, 'stream_out'))

  return (streams_in, stream_out)

def get_slave_devices(machine, config, section):
  machine.DEBUG('get_slave_devices: section=%s', section)

  input_device, output_device = None, None

  input_spec = config.get(section, 'input', None)
  output_spec = config.get(section, 'output', None)

  if input_spec is not None:
    backend_name, frontend_driver = input_spec.split(':')

    input_device = get_driver_creator(frontend_driver)(machine, machine.config, backend_name)

  if output_spec is not None:
    backend_name, frontend_driver = output_spec.split(':')

    output_device = get_driver_creator(frontend_driver)(machine, machine.config, backend_name)

  return (input_device, output_device)

class Terminal(DeviceFrontend):
  def __init__(self, machine, name, echo = False, *args, **kwargs):
    super(Terminal, self).__init__(machine, 'terminal', name)

    self._input = None
    self._output = None

    self._echo = echo
    self._input_read_u8_orig = None

  def _input_read_u8_echo(self, *args, **kwargs):
    c = self._input_read_u8_orig(*args, **kwargs)

    if c != 0xFF:
      self._output.write_u8(self._output.port, c)

    return c

  def _patch_echo(self, restore = False):
    D = self.machine.DEBUG

    D('%s._patch_echo: echo=%s, restore=%s', self.__class__.__name__, self._echo, restore)

    if restore is True and self._input_read_u8_orig is not None:
      self._input.read_u8, self._input_read_u8_orig = self._input_read_u8_orig, None

    elif self._echo is True:
      assert self._input is not None
      assert hasattr(self._input, 'read_u8')
      assert hasattr(self._output, 'write_u8')

      self._input_read_u8_orig, self._input.read_u8 = self._input.read_u8, self._input_read_u8_echo

    D('%s.patch_echo: input.read_u8=%s, orig_input.read_u8=%s', self.__class__.__name__, self._input.read_u8, self._input_read_u8_orig)

  def boot(self):
    super(Terminal, self).boot()

    # self._patch_echo()

  def halt(self):
    super(Terminal, self).halt()

    # self._patch_echo(restore = True)

class StreamIOTerminal(Terminal):
  def __init__(self, machine, name, input_device = None, output_device = None, *args, **kwargs):
    super(StreamIOTerminal, self).__init__(machine, name, *args, **kwargs)

    machine.DEBUG('%s: name=%s, input_device=%s, output_device=%s', self.__class__.__name__, name, input_device, output_device)

    self._input = input_device
    self._output = output_device

    self._streams_in = []
    self._stream_out = None

    self._input.master = self
    self._output.master = self

  def enqueue_input_stream(self, stream):
    self.machine.DEBUG('%s.enqueue_input_stream: stream=%r', self.__class__.__name__, stream)

    self._input.enqueue_stream(stream)

  def enqueue_streams(self, streams_in = None, stream_out = None):
    self.machine.DEBUG('%s.enqueue_streams: streams_in=%s, stream_out=%s', self.__class__.__name__, streams_in, stream_out)

    if streams_in is not None:
      streams_in = streams_in or []

      for stream in streams_in:
        self.enqueue_input_stream(stream)

      self._streams_in = streams_in

    if stream_out is not None:
      self._stream_out = stream_out
      self._output.set_output(stream_out)

  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = get_slave_devices(machine, config, section)

    term = StreamIOTerminal(machine, section, input_device = input_device, output_device = output_device, echo = config.getbool(section, 'echo', False))

    streams_in, stream_out = parse_io_streams(machine, config, section)
    term.enqueue_streams(streams_in = streams_in, stream_out = stream_out)

    return term

  def boot(self):
    super(StreamIOTerminal, self).boot()

    self._input.boot()
    self._output.boot()

    self.machine.tenh('hid: basic terminal (%s, %s)', self._input.name, self._output.name)

  def halt(self):
    super(StreamIOTerminal, self).halt()

    self._input.halt()
    self._output.halt()

    for stream in self._streams_in:
      stream.close()

    if self._stream_out is not None:
      self._stream_out.flush()
      self._stream_out.close()

    self.machine.DEBUG('Standard terminal halted.')

class StandardIOTerminal(StreamIOTerminal):
  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = get_slave_devices(machine, config, section)

    term = StandardIOTerminal(machine, section, input_device = input_device, output_device = output_device)
    term.enqueue_streams(streams_in = [InputStream.create(machine, '<stdin>')], stream_out = OutputStream.create(machine, '<stdout>'))

    return term

class StandalonePTYTerminal(StreamIOTerminal):
  def __init__(self, *args, **kwargs):
    super(StandalonePTYTerminal, self).__init__(*args, **kwargs)

    self.pttys = None

  @staticmethod
  def create_from_config(machine, config, section):
    input_device, output_device = get_slave_devices(machine, config, section)

    term = StandalonePTYTerminal(machine, section, input_device = input_device, output_device = output_device, echo = config.getbool(section, 'echo', False))

    streams_in, stream_out = parse_io_streams(machine, config, section)
    term.enqueue_streams(streams_in = streams_in, stream_out = stream_out)

    return term

  def boot(self):
    self.machine.DEBUG('StandalonePTYTerminal.boot')

    Terminal.boot(self)

    if self.pttys is not None:
      return

    import pty

    pttys = pty.openpty()

    self.machine.DEBUG('  set I/O stream: pttys=%s', pttys)

    self.enqueue_streams(streams_in = [InputStream.create(self.machine, pttys[0])], stream_out = OutputStream.create(self.machine, pttys[0]))

    self.terminal_device = os.ttyname(pttys[1]) if pttys else '/dev/unknown'

    self._input.boot()
    self._output.boot()

    self.pttys = pttys

    self.machine.tenh('hid: pty terminal (%s, %s), dev %s', self._input.name, self._output.name, self.terminal_device)

  def halt(self):
    self.machine.DEBUG('StandalonePTYTerminal.halt')

    Terminal.halt(self)

    if self.pttys is None:
      return

    self._input.halt()
    self._output.halt()

    try:
      os.close(self.pttys[1])
      os.close(self.pttys[0])

      self.pttys = None
      self.terminal_device = None

    except Exception:
      self.machine.EXCEPTION('Exception raised while closing PTY')

    self.machine.DEBUG('StandalonePTYTerminal: halted')



import ast
import _ast
import imp
import inspect
import os
import sys

from six import exec_, print_

if os.getenv('VMDEBUG_PATCH', None) == 'yes':
  def debug(s):
    print_('PATCH:', s)

else:
  def debug(s):
    pass

def exec_f(object_, globals_ = None, locals_ = None):
  if not globals_ and not locals_:
    frame = inspect.stack()[1][0]
    globals_ = frame.f_globals
    locals_ = frame.f_locals

  elif globals_ and not locals_:
    locals_ = globals_

  exec_(object_, globals_, locals_)

class RemoveLoggingVisitor(ast.NodeTransformer):
  def visit_Expr(self, node):
    self.generic_visit(node)

    if type(node.value) != ast.Call:
      return node

    if not hasattr(node.value, 'func'):
      return node

    f = node.value.func

    if hasattr(f, "id"):
      if f.id == 'debug':
        debug('removing %s - it is "debug" call' % node)
        return None

      if f.id == 'DEBUG':
        debug('removing %s - it is "DEBUG" call' % node)
        return None

      if f.id == 'log_cpu_core_state':
        debug('removing %s - it is log_cpu_core_state call' % node)
        return None

      if f.id == 'D':
        debug('removing %s - it is D call' % node)
        return None

      return node

    else:
      if hasattr(f, 'attr') and (f.attr == 'DEBUG' or f.attr == 'debug'):
        debug('removing %s - it is DEBUG call' % node)
        return None

    return node

  def visit_If(self, node):
    self.generic_visit(node)

    # if __debug__:
    if isinstance(node.test, _ast.Name) and node.test.id == '__debug__':
      debug('replacing %s - it is __debug__ if' % node)
      debug('  orelse: %s' % node.orelse)

      return node.orelse

    if not node.body:
      debug('removing %s - it is empty if body' % node)
      return None

    return node

  def visit_For(self, node):
    self.generic_visit(node)

    if not node.body:
      debug('removing %s - it is empty for loop' % node)
      return None

    return node

class ModuleLoader(object):
  def __init__(self, fullpath):
    self.fullpath = fullpath

  def get_source(self, path):
    with open(path, 'rb') as f:
      source = f.read()

    return source

  def get_code(self, fullname):
    pkg = self.fullpath.endswith('__init__.py')

    debug('loading module: fullname=%s' % fullname)

    code_str = self.get_source(self.fullpath)
    code_tree = ast.parse(code_str)

    # Modify AST
    if fullname.startswith('ducky'):
      visitor = RemoveLoggingVisitor()
      new_code_tree = visitor.visit(code_tree)

    else:
      new_code_tree = code_tree

    code = compile(new_code_tree, self.fullpath, 'exec')

    return (pkg, code)

  def load_module(self, fullname):
    pkg, code = self.get_code(fullname)

    mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
    mod.__file__ = self.fullpath
    mod.__loader__ = self

    if pkg:
      mod.__path__ = [os.path.dirname(self.fullpath)]

    exec_f(code, mod.__dict__)

    return mod

class Importer(object):
  def find_module(self, fullname, path = None):
    path = path or sys.path

    for directory in path:
      loader = self.loader_for_path(directory, fullname)
      if loader:
        return loader

  def loader_for_path(self, directory, fullname):
    debug('module %s from %s' % (fullname, directory))

    module_path = os.path.join(directory, fullname.split('.')[-1]) + ".py"
    if os.path.exists(module_path):
      loader = ModuleLoader(module_path)
      return loader

    package_path = os.path.join(directory, fullname.split('.')[-1], '__init__.py')
    if os.path.exists(package_path):
      loader = ModuleLoader(package_path)
      return loader

if '-d' not in sys.argv:
  sys.meta_path.insert(0, Importer())



"""
SimpleVGA is very basic implementation of VGA-like device, with text
and graphic modes.
"""

import array
import enum
import functools

from six.moves import range

from ctypes import c_ushort, LittleEndianStructure

from . import Device, IOProvider
from ..errors import InvalidResourceError
from ..mm import PAGE_SIZE, ExternalMemoryPage, addr_to_page, u8_t
from ..util import sizeof_fmt, F, UINT16_FMT, UINT32_FMT
from ..reactor import RunInIntervalTask
from ..streams import OutputStream

#: Default address of command port
DEFAULT_PORT_RANGE = 0x3F0
#: Default memory size, in bytes
DEFAULT_MEMORY_SIZE = 64 * 1024
#: Default number of memory banks
DEFAULT_MEMORY_BANKS = 8


class SimpleVGACommands(enum.IntEnum):
  RESET          = 0x8001
  REFRESH        = 0x0002

  GRAPHIC        = 0x0020
  COLS           = 0x0021
  ROWS           = 0x0022
  DEPTH          = 0x0023

  MEMORY_BANK_ID = 0x0030


class Mode(object):
  def __init__(self, _type, width, height, depth):
    self.type = _type
    self.width = width
    self.height = height
    self.depth = depth

    self.required_memory = self.width * self.height * (self.depth if self.type == 't' else self.depth // 8)

  def __cmp__(self, other):
    return self.type == other.type and self.width == other.width and self.height == other.height and self.depth == other.depth

  def __eq__(self, other):
    return self.__cmp__(other)

  def __repr__(self):
    return self.to_string()

  @classmethod
  def from_string(cls, s):
    """
    Create ``Mode`` object from its string representation. It's a comma-separated
    list of for items:

    +--------------+----------+-----------------+-----------------------+-----------------+
    |              | ``type`` | ``width``       | ``height``            | ``depth``       |
    +--------------+----------+-----------------+-----------------------+-----------------+
    | Text mode    | ``t``    | chars per line  | lines on screen       | bytes per char  |
    +--------------+----------+-----------------+-----------------------+-----------------+
    | Graphic mode | ``g``    | pixels per line | pixel lines on screen | bites per pixel |
    +--------------+----------+-----------------+-----------------------+-----------------+
    """

    t, c, r, b = s.strip().split(',')
    return cls(t, int(c.strip()), int(r.strip()), int(b.strip()))

  def to_string(self):
    return F('({type}, {width}, {height}, {depth})', type = self.type, width = self.width, height = self.height, depth = self.depth)

  def to_pretty_string(self):
    return F('{type}, {cols}x{rows} {entities}, {memory_per_entity} {memory_label}',
             type = 'text' if self.type == 't' else 'graphic',
             cols = self.width,
             rows = self.height,
             entities = 'chars' if self.type == 't' else 'pixels',
             memory_per_entity = self.depth if self.type == 't' else self.depth * 8,
             memory_label = 'bytes per char' if self.type == 't' else 'bits color depth'
             )

#: Default list of available modes
DEFAULT_MODES = [
  Mode('g', 320, 200, 1),
  Mode('t', 80, 25, 2),
  Mode('t', 80, 25, 1)
]
#: Default boot mode
DEFAULT_BOOT_MODE = Mode('t', 80, 25, 1)

class Char(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('codepoint', c_ushort, 7),
    ('unused',    c_ushort, 1),
    ('fg',        c_ushort, 4),
    ('bg',        c_ushort, 3),
    ('blink',     c_ushort, 1)
  ]

  def __repr__(self):
    return '<Char: cp=%s, fg=%s, bg=%s, bl=%s>' % (self.codepoint, self.fg, self.bg, self.blink)

  def to_u8(self):
    return (u8_t(self.codepoint | self.unused << 7).value, u8_t(self.fg | self.bg << 4 | self.blink << 7).value)

  @staticmethod
  def from_u8(l, h):
    c = Char()
    c.codepoint = l & 0x7F
    c.fg = h & 0x0F
    c.bg = (h >> 4) & 0x07
    c.blink = h >> 7
    return c

  @staticmethod
  def from_u16(u):
    return Char.from_u8(u & 0x00FF, u >> 8)

class DisplayRefreshTask(RunInIntervalTask):
  def __init__(self, display):
    super(DisplayRefreshTask, self).__init__(200, self.on_tick)

    self.display = display
    self.first_tick = True

    self.write = display.stream_out.write

  def on_tick(self, task):
    self.display.machine.DEBUG('Display: refresh display')

    gpu = self.display.gpu
    mode = gpu.active_mode

    palette_fg = [30, 34, 32, 36, 31, 35, 31, 37,  90,  94,  92,  96,  91,  95, 33,  97]
    palette_bg = [40, 44, 42, 46, 41, 45, 41, 47, 100, 104, 102, 106, 101, 105, 43, 107]

    if mode.type == 't':
      screen = []

      if mode.depth not in (1, 2):
        self.display.machine.WARN(F('Unhandled character depth: mode={mode}', mode = self.active_mode))
        return

      for row in range(0, mode.height):
        line = []

        for col in range(0, mode.width):
          if mode.depth == 1:
            c = Char.from_u8(gpu.memory[row * mode.width + col], 0)
            c.fg = 15
            c.bg = 0

          else:
            c = Char.from_u8(gpu.memory[(row * mode.width + col) * 2], gpu.memory[(row * mode.width + col) * 2 + 1])

          self.display.machine.DEBUG('  c=%s', c)

          char = F('\033[{blink:d};{fg:d};{bg:d}m{char}\033[0m',
                   blink = 5 if c.blink == 1 else 0,
                   fg = palette_fg[c.fg],
                   bg = palette_bg[c.bg],
                   char = ' ' if c.codepoint == 0 else chr(c.codepoint)
                   )
          line.append(char)

        screen.append(''.join(line))

      def __line(l):
        return [ord(c) for c in l] + [ord('\n')]

      if self.first_tick:
        self.first_tick = False
      else:
        self.write(__line(F('\033[{rows:d}F', rows = mode.height + 3)))

      self.write(__line('-' * mode.width))
      for line in screen:
        self.write(__line(line))
      self.write(__line('-' * mode.width))

    else:
      self.display.machine.WARN(F('Unhandled gpu mode: mode={mode}', mode = gpu.active_mode))

class Display(Device):
  def __init__(self, machine, name, gpu = None, stream_out = None, *args, **kwargs):
    super(Display, self).__init__(machine, 'display', name, *args, **kwargs)

    self.gpu = gpu
    self.stream_out = stream_out

    self.gpu.master = self

    self.refresh_task = DisplayRefreshTask(self)

  @staticmethod
  def get_slave_gpu(machine, config, section):
    gpu_name = config.get(section, 'gpu', None)
    gpu_device = machine.get_device_by_name(gpu_name)

    if not gpu_name or not gpu_device:
      raise InvalidResourceError(F('Unknown GPU device: gpu={name}', name = gpu_name))

    return gpu_device

  @staticmethod
  def create_from_config(machine, config, section):
    gpu = Display.get_slave_gpu(machine, config, section)
    stream_out =  OutputStream.create(machine, config.get(section, 'stream_out', '<stdout>'))

    return Display(machine, section, gpu = gpu, stream_out = stream_out)

  def boot(self):
    self.machine.DEBUG('Display.boot')

    super(Display, self).boot()

    self.gpu.boot()
    self.machine.reactor.add_task(self.refresh_task)
    self.machine.reactor.task_runnable(self.refresh_task)

    self.machine.tenh(F('display: generic {name} connected to gpu {gpu}, output stream {stream}', name = self.name, gpu = self.gpu.name, stream = self.stream_out))

  def halt(self):
    self.machine.DEBUG('Display.halt')

    super(Display, self).halt()

    self.machine.reactor.remove_task(self.refresh_task)
    self.gpu.halt()

    self.machine.DEBUG('Display: halted')


class SimpleVGAMemoryPage(ExternalMemoryPage):
  """
  Memory page handling MMIO of sVGA device.

  :param ducky.devices.svga.SimpleVGA dev: sVGA device this page belongs to.
  """

  def __init__(self, dev, *args, **kwargs):
    super(SimpleVGAMemoryPage, self).__init__(*args, **kwargs)

    self.dev = dev

  def get(self, offset):
    return self.data[self.dev.bank_offsets[self.dev.active_bank] + self.offset + offset]

  def put(self, offset, b):
    self.data[self.dev.bank_offsets[self.dev.active_bank] + self.offset + offset] = b

class SimpleVGA(IOProvider, Device):
  """
  SimpleVGA is very basic implementation of VGA-like device, with text
  and graphic modes.

  It has its own graphic memory ("buffer"), split into several banks of
  the same size. Always only one bank can be directly accessed, by having
  it mapped into CPU's address space.

  :param ducky.machine.Machine machine: machine this device belongs to.
  :param string name: name of this device.
  :param u16 port: address of the command port.
  :param int memory_size: size of graphic memory.
  :param u24 memory_address: address of graphic memory - to this address is
    graphic buffer mapped. Must be specified, there is no default value.
  :param int memory_banks: number of memory banks.
  :param list modes: list of :py:class:`ducky.devices.svga.Mode` objects,
    list of supported modes.
  :param tuple boot_mode: this mode will be set when device boots up.
  """

  def __init__(self, machine, name, port = None, memory_size = None, memory_address = None, memory_banks = None, modes = None, boot_mode = None, *args, **kwargs):
    if memory_address is None:
      raise InvalidResourceError('sVGA device memory address must be specified explicitly')

    if memory_address % PAGE_SIZE:
      raise InvalidResourceError('sVGA device memory address must be page-aligned')

    super(SimpleVGA, self).__init__(machine, 'gpu', name, *args, **kwargs)

    self.port = port or DEFAULT_PORT_RANGE
    self.ports = [port, port + 0x0001]

    self.memory_size = memory_size or DEFAULT_MEMORY_SIZE
    self.memory_address = memory_address
    self.memory_banks = memory_banks or DEFAULT_MEMORY_BANKS
    self.modes = modes or DEFAULT_MODES

    if self.memory_size % PAGE_SIZE:
      raise InvalidResourceError('sVGA device memory size must be page-aligned')

    if (self.memory_size // self.memory_banks) % PAGE_SIZE:
      raise InvalidResourceError('sVGA device memory bank size must be page-aligned')

    self.active_mode = None
    self.boot_mode = boot_mode or DEFAULT_BOOT_MODE

    if self.boot_mode not in self.modes:
      raise InvalidResourceError(F('Boot mode not available: boot_mode={mode}, modes={modes}', mode = self.boot_mode, modes = self.modes))

    for mode in self.modes:
      if mode.required_memory > self.memory_size:
        raise InvalidResourceError(F('Not enough memory for mode: mode={mode}, required={bytes_required:d} bytes, available={bytes_available:d} bytes', mode = mode, bytes_required = mode.required_memory, bytes_available = self.memory_size))

    self.memory = self.data = array.array('B', [0 for _ in range(0, self.memory_size)])
    self.bank_offsets = list(range(0, self.memory_size, self.memory_size // self.memory_banks))
    self.pages_per_bank = self.memory_size // PAGE_SIZE // self.memory_banks

    self.machine.DEBUG(F('sVGA: memory-size={memory_size:d}, memory-banks={memory_banks:d}, offsets=[{bank_offsets}], pages-per-bank={pages_per_bank:d}, address={address:W}', memory_size = self.memory_size, memory_banks = self.memory_banks, bank_offsets = ', '.join([UINT32_FMT(o) for o in self.bank_offsets]), pages_per_bank = self.pages_per_bank, address = self.memory_address))

  @staticmethod
  def create_from_config(machine, config, section):
    _getint = functools.partial(config.getint, section)

    def parse_mode(m):
      t, c, r, b = m.strip().split(',')
      return (t, int(c.strip()), int(r.strip()), int(b.strip()))

    modes = config.get(section, 'modes', None)
    if modes is not None:
      modes = [Mode.from_string(m) for m in modes.split(';')]

    boot_mode = config.get(section, 'boot-mode', None)
    if boot_mode is not None:
      boot_mode = Mode.from_string(boot_mode)

    return SimpleVGA(machine, section,
                     port = _getint('port', DEFAULT_PORT_RANGE),
                     memory_size = _getint('memory-size', DEFAULT_MEMORY_SIZE),
                     memory_address = _getint('memory-address', None),
                     memory_banks = _getint('memory-banks', DEFAULT_MEMORY_BANKS),
                     modes = modes,
                     boot_mode = boot_mode)

  def __repr__(self):
    return F('sVGA adapter {name} ({memory_size} VRAM in {memory_banks:d} banks at {memory_address}, control [{ports}]; {mode} mode)',
             name = self.name,
             memory_size = sizeof_fmt(self.memory_size),
             memory_banks = self.memory_banks,
             memory_address = UINT32_FMT(self.memory_address),
             ports = ', '.join([UINT16_FMT(port) for port in self.ports]),
             mode = self.active_mode.to_pretty_string() if self.active_mode is not None else (self.boot_mode.to_pretty_string() + ' boot')
             )

  def reset(self):
    self.state = None
    self.active_mode = None
    self.active_bank = 0

    for pg in self.pages:
      pg.clear()

  def set_mode(self, mode):
    self.active_mode = mode

  def boot(self):
    self.machine.DEBUG('SimpleVGA.boot')

    for port in self.ports:
      self.machine.register_port(port, self)

    self.pages = []
    pages_start = addr_to_page(self.memory_address)

    for i in range(pages_start, pages_start + self.pages_per_bank):
      pg = SimpleVGAMemoryPage(self, self.machine.memory, i, self.memory, offset = (i - pages_start) * PAGE_SIZE)
      self.machine.memory.register_page(pg)
      self.pages.append(pg)

    self.reset()
    self.set_mode(self.boot_mode)

    self.machine.tenh(F('gpu: {gpu}', gpu = self))

  def halt(self):
    self.machine.DEBUG('SimpleVGA.halt')

    for pg in self.pages:
      self.machine.memory.unregister_page(pg)

    self.pages = []

    for port in self.ports:
      self.machine.unregister_port(port)

    self.machine.DEBUG('SimpleVGA: halted')

  def read_u16(self, port):
    self.machine.DEBUG(F('{method}.read_u16: port={port:S}', method = self.__class__.__name__, port = port))

    if port != self.ports[1]:
      raise InvalidResourceError('Unable to read from command register')

    if self.state == SimpleVGACommands.GRAPHIC:
      self.state = None
      return 1 if self.active_mode.type == 'g' else 0

    if self.state == SimpleVGACommands.COLS:
      self.state = None
      return self.active_mode.width

    if self.state == SimpleVGACommands.ROWS:
      self.state = None
      return self.active_mode.height

    if self.state == SimpleVGACommands.DEPTH:
      self.state = None
      return self.active_mode.depth

    if self.state == SimpleVGACommands.MEMORY_BANK_ID:
      self.state = None
      return self.active_bank

    raise InvalidResourceError(F('Invalid internal state: state={state}', state = self.state))

  def write_u16(self, port, value):
    self.machine.DEBUG(F('{method}.write_u16: port={port:S}, value={value:S}', method = self.__class__.__name__, port = port, value = value))

    if port not in self.ports:
      raise InvalidResourceError(F('Unhandled port: port={port:S}', port = port))

    if self.ports.index(port) == 0:
      # command port

      if value == SimpleVGACommands.RESET:
        self.reset()
        return

      if value == SimpleVGACommands.REFRESH:
        self.master.refresh_task.on_tick(None)
        return

      self.state = value

    else:
      # data port

      if self.state == SimpleVGACommands.MEMORY_BANK_ID:
        if not 0 <= value < self.memory_banks:
          raise InvalidResourceError(F('Memory bank out of range: bank={bank:d}', bank = value))

        self.active_bank = value
        self.state = None

      else:
        raise InvalidResourceError(F('Invalid internal state: state={state}', state = self.state))



import enum
import mmap

from six.moves import range

from .. import cpu

from ..mm import u8_t, u16_t, u32_t, UINT32_FMT
from ..util import BinaryFile, StringTable, align, Flags, str2bytes, bytes2str
from ctypes import LittleEndianStructure, sizeof

class SectionFlagsEncoding(LittleEndianStructure):
  _fields_ = [
    ('readable',         u8_t, 1),
    ('writable',         u8_t, 1),
    ('executable',       u8_t, 1),
    ('loadable',         u8_t, 1),
    ('bss',              u8_t, 1),
    ('mmapable',         u8_t, 1),
    ('globally_visible', u8_t, 1),
  ]

class SectionFlags(Flags):
  _encoding = SectionFlagsEncoding
  _flags = [field[0] for field in SectionFlagsEncoding._fields_]
  _labels = 'RWELBMG'

class SectionTypes(enum.IntEnum):
  UNKNOWN = 0
  TEXT    = 1
  DATA    = 2
  SYMBOLS = 3
  STRINGS = 4
  RELOC   = 5

SECTION_TYPES = [
  'UNKNOWN', 'TEXT', 'DATA', 'SYMBOLS', 'STRINGS', 'RELOC'
]

class SymbolDataTypes(enum.IntEnum):
  INT      = 0
  SHORT    = 1
  CHAR     = 2
  BYTE     = 3
  STRING   = 4
  ASCII    = 5
  FUNCTION = 6
  UNKNOWN  = 7

SYMBOL_DATA_TYPES = 'ISCBTAFU'

class FileFlagsEncoding(LittleEndianStructure):
  _fields_ = [
    ('mmapable', u16_t, 1)
  ]

class FileFlags(Flags):
  _encoding = FileFlagsEncoding
  _flags = [field[0] for field in FileFlagsEncoding._fields_]
  _labels = 'M'

class FileHeader(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('magic',    u16_t),
    ('version',  u16_t),
    ('flags',    FileFlagsEncoding),
    ('sections', u16_t)
  ]

class SectionHeader(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('index',     u8_t),
    ('name',      u32_t),
    ('type',      u8_t),
    ('flags',     SectionFlagsEncoding),
    ('padding',   u8_t),
    ('base',      u32_t),
    ('items',     u32_t),
    ('data_size', u32_t),
    ('file_size', u32_t),
    ('offset',    u32_t)
  ]

  def __repr__(self):
    return '<SectionHeader: index={}, name={}, type={}, flags={}, base={}, items={}, data_size={}, file_size={}, offset={}>'.format(self.index, self.name, self.type, SectionFlags.from_encoding(self.flags).to_string(), UINT32_FMT(self.base), self.items, self.data_size, self.file_size, self.offset)

class SymbolFlagsEncoding(LittleEndianStructure):
  _fields_ = [
    ('globally_visible', u16_t, 1)
  ]

class SymbolFlags(Flags):
  _encoding = SymbolFlagsEncoding
  _flags = [field[0] for field in SymbolFlagsEncoding._fields_]
  _labels = 'G'

class SymbolEntry(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('flags',        SymbolFlagsEncoding),
    ('name',         u32_t),
    ('address',      u32_t),
    ('size',         u32_t),
    ('section',      u8_t),
    ('type',         u8_t),
    ('filename',     u32_t),
    ('lineno',       u32_t)
  ]

  def __repr__(self):
    return '<SymbolEntry: section={}, name={}, type={}, addr={}, flags={}>'.format(self.section, self.name, SYMBOL_DATA_TYPES[self.type], UINT32_FMT(self.address), SymbolFlags.from_encoding(self.flags).to_string())

class RelocFlagsEncoding(LittleEndianStructure):
  _fields_ = [
    ('relative',     u16_t, 1),
    ('inst_aligned', u16_t, 1)
  ]

class RelocFlags(Flags):
  _encoding = RelocFlagsEncoding
  _flags = [field[0] for field in RelocFlagsEncoding._fields_]
  _labels = 'RI'

class RelocEntry(LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    ('flags',         RelocFlagsEncoding),
    ('name',          u32_t),
    ('patch_add',     u32_t),
    ('patch_address', u32_t),
    ('patch_section', u8_t),
    ('patch_offset',  u8_t),
    ('patch_size',    u8_t)
  ]

  def __repr__(self):
    return '<RelocEntry: flags=%s, name=%s, section=%s, address=%s, offset=%s, size=%s, add=%s>' % (RelocFlags.from_encoding(self.flags).to_string(), self.name, self.patch_section, UINT32_FMT(self.patch_address), self.patch_offset, self.patch_size, self.patch_add)

SECTION_ITEM_SIZE = [
  0, 4, sizeof(u8_t), sizeof(SymbolEntry)
]

class File(BinaryFile):
  MAGIC = 0xDEAD
  VERSION = 1

  @staticmethod
  def open(*args, **kwargs):
    return BinaryFile.do_open(*args, klass = File, **kwargs)

  def setup(self):
    self.__header = None
    self.__sections = []

    self.string_table = StringTable()

    self.symbols = None

  def create_header(self):
    self.__header = FileHeader()
    self.__header.magic = self.MAGIC
    self.__header.version = self.VERSION
    self.__header.sections = 0

    return self.__header

  def create_section(self):
    header = SectionHeader()
    header.index = len(self.__sections)
    self.__sections.append((header, []))
    return header

  def set_content(self, header, content):
    self.__sections[header.index] = (header, content)
    header.items = len(content)

  def get_header(self):
    return self.__header

  def sections(self):
    return (self.get_section(i) for i in range(0, self.get_header().sections))

  def get_section(self, i):
    return self.__sections[i]

  def get_section_by_name(self, name):
    for i in range(0, self.get_header().sections):
      header, content = self.get_section(i)

      if self.string_table.get_string(header.name) == name:
        return header, content

    from ..mm import MalformedBinaryError
    raise MalformedBinaryError('Unknown section named "{}"'.format(name))

  def load_symbols(self):
    self.symbols = {}

    for header, content in self.__sections:
      if header.type != SectionTypes.SYMBOLS:
        continue

      for symbol in content:
        self.symbols[self.string_table.get_string(symbol.name)] = symbol

  def load(self):
    self.seek(0)

    self.DEBUG('load: loading headers')

    self.__header = self.read_struct(FileHeader)

    if self.__header.magic != self.MAGIC:
      self.ERROR('load: magic cookie not recognized!')
      from ..mm import MalformedBinaryError
      raise MalformedBinaryError('Magic cookie not recognized!')

    for i in range(0, self.__header.sections):
      self.__sections.append((self.read_struct(SectionHeader), []))

    for i in range(0, self.__header.sections):
      header, content = self.__sections[i]

      self.DEBUG('load: loading section #%i', header.index)

      self.seek(header.offset)

      if header.type == SectionTypes.STRINGS:
        self.string_table.buff = bytes2str(self.read(header.file_size))

      else:
        if header.type == SectionTypes.DATA:
          count = header.data_size
          st_class = u8_t

        elif header.type == SectionTypes.SYMBOLS:
          count = header.items
          st_class = SymbolEntry

        elif header.type == SectionTypes.TEXT:
          count = header.items
          st_class = u32_t

        elif header.type == SectionTypes.RELOC:
          count = header.items
          st_class = RelocEntry

        else:
          from ..mm import MalformedBinaryError
          raise MalformedBinaryError('Unknown section header type {}'.format(header.type))

        for _ in range(0, count):
          content.append(self.read_struct(st_class))

  def save(self):
    self.seek(0)

    self.__header.sections = len(self.__sections)

    offset = sizeof(FileHeader) + sizeof(SectionHeader) * self.__header.sections

    if self.__header.flags.mmapable == 1:
      offset = align(mmap.PAGESIZE, offset)

    for i in range(0, len(self.__sections)):
      header, content = self.__sections[i]

      header.offset = offset
      self.DEBUG('set section %s offset to %s', self.string_table.get_string(header.name), offset)

      if header.type == SectionTypes.STRINGS:
        header.data_size = header.file_size = len(self.string_table.buff)

      if header.flags.bss != 1:
        offset += header.file_size
        self.DEBUG('extending offset by %s to %s', header.file_size, offset)

        if self.__header.flags.mmapable == 1:
          offset = align(mmap.PAGESIZE, offset)
          self.DEBUG('offset aligned to %s', offset)

      self.DEBUG(str(header))
      self.DEBUG('')

    self.DEBUG('save: saving headers')

    self.write_struct(self.__header)

    for i in range(0, len(self.__sections)):
      self.write_struct(self.__sections[i][0])

    for i in range(0, len(self.__sections)):
      header, content = self.__sections[i]

      self.DEBUG('save: saving section %s: %s', self.string_table.get_string(header.name), header)

      self.seek(header.offset)

      if header.type == SectionTypes.STRINGS:
        self.DEBUG('write: %s', self.string_table.buff)
        self.write(str2bytes(self.string_table.buff))

      elif header.flags.bss == 1:
        self.DEBUG('BSS section - dont write out any slots')

      else:
        for item in content:
          if type(item) == cpu.assemble.SpaceSlot:
            self.DEBUG('write_space: %s bytes', item.size)
            self.write(str2bytes('\x00' * item.size))

          else:
            self.write_struct(item)

        if header.data_size != header.file_size:
          self.DEBUG('write after-section padding of %s bytes', header.file_size - header.data_size)
          self.write(str2bytes('\x00' * (header.file_size - header.data_size)))
